# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>
#            Matt Prahl <mprahl@redhat.com>
#            Jan Kaluza <jkaluza@redhat.com>
import re
import time
import shutil
import tempfile
import os
from multiprocessing.dummy import Pool as ThreadPool
from datetime import datetime

import kobo.rpmlib
import requests
from gi.repository import GLib

import module_build_service.scm
import module_build_service.resolver

from module_build_service import conf, db, log, models, Modulemd
from module_build_service.errors import (
    ValidationError, UnprocessableEntity, Forbidden, Conflict)
from module_build_service import glib
from .mse import generate_expanded_mmds
from module_build_service.utils.ursine import record_stream_collision_modules


def record_filtered_rpms(mmd):
    """Record filtered RPMs that should not be installed into buildroot

    These RPMs are filtered:

    * Reads the mmd["xmd"]["buildrequires"] and extends it with "filtered_rpms"
      list containing the NVRs of filtered RPMs in a buildrequired module.

    * Every base module could have stream collision modules, whose built RPMs
      should be filtered out to avoid collision. The filtered out RPMs will be
      stored into the base module as well.

    :param Modulemd mmd: Modulemd that will be built next.
    :rtype: Modulemd.Module
    :return: Modulemd extended with the "filtered_rpms" in XMD section.
    """
    new_buildrequires = {}
    for req_name, req_data in mmd.get_xmd()["mbs"]["buildrequires"].items():
        _record_filtered_rpms(req_name, req_data)
        if req_name in conf.base_module_names and 'stream_collision_modules' in req_data:
            _record_ursine_rpms(req_data)
        new_buildrequires[req_name] = req_data

    # Replace the old buildrequires with new ones.
    xmd = glib.from_variant_dict(mmd.get_xmd())
    xmd["mbs"]["buildrequires"] = new_buildrequires
    mmd.set_xmd(glib.dict_values(xmd))
    return mmd


def _record_filtered_rpms(req_name, req_data):
    """Store filtered RPMs by a buildrequire module

    This methods looks for key ``filtered_rpms`` in a buildrequired module's
    metadata. If there is, nothing is done, just keep it unchanged, which case
    could be a module is resubmitted or a local build.

    Otherwise, ``_record_filtered_rpms`` will get module's RPMs listed in
    filter section, then store them with the key into metadata in place.

    :param str req_name: name of a buildrequired module.
    :param dict req_data: a mapping containing metadata of the buildrequired
        module. A pair of ``req_name`` and ``req_data`` is usually the one of
        xmd.mbs.buildrequires, which are stored during the module stream
        expansion process. At least, ``req_data`` must have keys stream,
        version and context. Key ``filtered_rpms`` will be added as a list of
        RPMs N-E:V-R.
    """
    # Imported here to allow import of utils in GenericBuilder.
    from module_build_service.builder import GenericBuilder

    # In case this is module resubmit or local build, the filtered_rpms
    # will already be there, so there is no point in generating them again.
    if "filtered_rpms" in req_data:
        return

    resolver = module_build_service.resolver.GenericResolver.create(conf)

    # We can just get the first modulemd data from result right here thanks to
    # strict=True, so in case the module cannot be found, get_module_modulemds
    # raises an exception.
    req_mmd = resolver.get_module_modulemds(
        req_name, req_data["stream"], req_data["version"], req_data["context"], True)[0]

    # Find out the particular NVR of filtered packages
    filtered_rpms = []
    rpm_filter = req_mmd.get_rpm_filter()
    if rpm_filter and rpm_filter.get():
        rpm_filter = rpm_filter.get()
        built_nvrs = GenericBuilder.backends[conf.system].get_built_rpms_in_module_build(
            req_mmd)
        for nvr in built_nvrs:
            parsed_nvr = kobo.rpmlib.parse_nvr(nvr)
            if parsed_nvr["name"] in rpm_filter:
                filtered_rpms.append(nvr)
    req_data["filtered_rpms"] = filtered_rpms


def _record_ursine_rpms(req_data):
    """Store ursine RPMs

    This method handles key ``stream_collision_modules`` from a buildrequired
    module's metadata to find out all built RPMs and then store them with a new
    key ``ursine_rpms`` into metadata in place.

    :param dict req_data: a mapping containing metadata of the buildrequired
        module. At least, ``req_data`` must have keys stream, version and
        context. As a result, a new key ``filtered_ursine_rpms`` will be added
        with a list of RPMs N-E:V-R which are built for the found stream
        collision modules.
    """
    from module_build_service.builder.KojiModuleBuilder import KojiModuleBuilder
    resolver = module_build_service.resolver.GenericResolver.create(conf)

    # Key stream_collision_modules is not used after rpms are recorded, but
    # just keep it here in case it would be helpful in the future.
    modules_nsvc = req_data['stream_collision_modules']
    built_rpms = []

    koji_session = KojiModuleBuilder.get_session(conf, None)

    for nsvc in modules_nsvc:
        name, stream, version, context = nsvc.split(':')
        module = resolver._get_module(name, stream, version, context, strict=True)
        rpms = koji_session.listTaggedRPMS(module['koji_tag'], latest=True)[0]
        built_rpms.extend(
            kobo.rpmlib.make_nvr(rpm, force_epoch=True) for rpm in rpms
        )

    # In case there is duplicate NEVRs, ensure every NEVR is unique in the final list.
    # And, sometimes, sorted list of RPMs would be easier to read.
    built_rpms = sorted(set(built_rpms))
    req_data['ursine_rpms'] = built_rpms


def _scm_get_latest(pkg):
    try:
        # If the modulemd specifies that the 'f25' branch is what
        # we want to pull from, we need to resolve that f25 branch
        # to the specific commit available at the time of
        # submission (now).
        pkgref = module_build_service.scm.SCM(
            pkg.get_repository()).get_latest(pkg.get_ref())
    except Exception as e:
        log.exception(e)
        return {'error': "Failed to get the latest commit for %s#%s" % (
            pkg.get_repository(), pkg.get_ref())}

    return {
        'pkg_name': pkg.get_name(),
        'pkg_ref': pkgref,
        'error': None
    }


def format_mmd(mmd, scmurl):
    """
    Prepares the modulemd for the MBS. This does things such as replacing the
    branches of components with commit hashes and adding metadata in the xmd
    dictionary.
    :param mmd: the Modulemd.Module object to format
    :param scmurl: the url to the modulemd
    """
    # Import it here, because SCM uses utils methods and fails to import
    # them because of dep-chain.
    from module_build_service.scm import SCM

    xmd = glib.from_variant_dict(mmd.get_xmd())
    if 'mbs' not in xmd:
        xmd['mbs'] = {}
    if 'scmurl' not in xmd['mbs']:
        xmd['mbs']['scmurl'] = scmurl or ''
    if 'commit' not in xmd['mbs']:
        xmd['mbs']['commit'] = ''

    # If module build was submitted via yaml file, there is no scmurl
    if scmurl:
        scm = SCM(scmurl)
        # If a commit hash is provided, add that information to the modulemd
        if scm.commit:
            # We want to make sure we have the full commit hash for consistency
            if SCM.is_full_commit_hash(scm.scheme, scm.commit):
                full_scm_hash = scm.commit
            else:
                full_scm_hash = scm.get_full_commit_hash()

            xmd['mbs']['commit'] = full_scm_hash
        # If a commit hash wasn't provided then just get the latest from master
        else:
            xmd['mbs']['commit'] = scm.get_latest()

    if mmd.get_rpm_components() or mmd.get_module_components():
        if 'rpms' not in xmd['mbs']:
            xmd['mbs']['rpms'] = {}
        # Add missing data in RPM components
        for pkgname, pkg in mmd.get_rpm_components().items():
            if pkg.get_repository() and not conf.rpms_allow_repository:
                raise Forbidden(
                    "Custom component repositories aren't allowed.  "
                    "%r bears repository %r" % (pkgname, pkg.get_repository()))
            if pkg.get_cache() and not conf.rpms_allow_cache:
                raise Forbidden(
                    "Custom component caches aren't allowed.  "
                    "%r bears cache %r" % (pkgname, pkg.cache))
            if not pkg.get_repository():
                pkg.set_repository(conf.rpms_default_repository + pkgname)
            if not pkg.get_cache():
                pkg.set_cache(conf.rpms_default_cache + pkgname)
            if not pkg.get_ref():
                pkg.set_ref('master')
            if pkg.get_arches().size() == 0:
                arches = Modulemd.SimpleSet()
                arches.set(conf.arches)
                pkg.set_arches(arches)

        # Add missing data in included modules components
        for modname, mod in mmd.get_module_components().items():
            if mod.get_repository() and not conf.modules_allow_repository:
                raise Forbidden(
                    "Custom module repositories aren't allowed.  "
                    "%r bears repository %r" % (modname, mod.get_repository()))
            if not mod.get_repository():
                mod.set_repository(conf.modules_default_repository + modname)
            if not mod.get_ref():
                mod.set_ref('master')

        # Check that SCM URL is valid and replace potential branches in pkg refs
        # by real SCM hash and store the result to our private xmd place in modulemd.
        pool = ThreadPool(20)
        try:
            pkg_dicts = pool.map(_scm_get_latest, mmd.get_rpm_components().values())
        finally:
            pool.close()

        err_msg = ""
        for pkg_dict in pkg_dicts:
            if pkg_dict["error"]:
                err_msg += pkg_dict["error"] + "\n"
            else:
                pkg_name = pkg_dict["pkg_name"]
                pkg_ref = pkg_dict["pkg_ref"]
                xmd['mbs']['rpms'][pkg_name] = {'ref': pkg_ref}
        if err_msg:
            raise UnprocessableEntity(err_msg)

    # Set the modified xmd back to the modulemd
    mmd.set_xmd(glib.dict_values(xmd))


def get_prefixed_version(mmd):
        """
        Return the prefixed version of the module based on the buildrequired base module stream.

        :param mmd: the Modulemd.Module object to format
        :return: the prefixed version
        :rtype: int
        """
        xmd = mmd.get_xmd()
        version = mmd.get_version()

        base_module_stream = None
        for base_module in conf.base_module_names:
            # xmd is a GLib Variant and doesn't support .get() syntax
            try:
                base_module_stream = xmd['mbs']['buildrequires'].get(
                    base_module, {}).get('stream')
                if base_module_stream:
                    # Break after finding the first base module that is buildrequired
                    break
            except KeyError:
                log.warning('The module\'s mmd is missing information in the xmd section')
                return version
        else:
            log.warning('This module does not buildrequire a base module ({0})'
                        .format(' or '.join(conf.base_module_names)))
            return version

        # The platform version (e.g. prefix1.2.0 => 010200)
        version_prefix = models.ModuleBuild.get_stream_version(base_module_stream, right_pad=False)

        if not version_prefix:
            log.warning('The "{0}" stream "{1}" couldn\'t be used to prefix the module\'s '
                        'version'.format(base_module, base_module_stream))
            return version

        new_version = int(str(version_prefix) + str(version))
        if new_version > GLib.MAXUINT64:
            log.warning('The "{0}" stream "{1}" caused the module\'s version prefix to be '
                        'too long'.format(base_module, base_module_stream))
            return version
        return new_version


def validate_mmd(mmd):
    """Validate module metadata

    If everything is ok, just keep quiet, otherwise error is raised for
    specific problem.

    :param mmd: modulemd object representing module metadata.
    :type mmd: Modulemd.Module
    :raises Forbidden: if metadata contains module repository but it is not
        allowed.
    """
    for modname, mod in mmd.get_module_components().items():
        if mod.get_repository() and not conf.modules_allow_repository:
            raise Forbidden(
                "Custom module repositories aren't allowed.  "
                "%r bears repository %r" % (modname, mod.get_repository()))


def merge_included_mmd(mmd, included_mmd):
    """
    Merges two modulemds. This merges only metadata which are needed in
    the `main` when it includes another module defined by `included_mmd`
    """
    included_xmd = glib.from_variant_dict(included_mmd.get_xmd())
    if 'rpms' in included_xmd['mbs']:
        xmd = glib.from_variant_dict(mmd.get_xmd())
        if 'rpms' not in xmd['mbs']:
            xmd['mbs']['rpms'] = included_xmd['mbs']['rpms']
        else:
            xmd['mbs']['rpms'].update(included_xmd['mbs']['rpms'])
    # Set the modified xmd back to the modulemd
    mmd.set_xmd(glib.dict_values(xmd))


def record_component_builds(mmd, module, initial_batch=1,
                            previous_buildorder=None, main_mmd=None, session=None):
    # Imported here to allow import of utils in GenericBuilder.
    import module_build_service.builder

    if not session:
        session = db.session

    # Format the modulemd by putting in defaults and replacing streams that
    # are branches with commit hashes
    format_mmd(mmd, module.scmurl)

    # When main_mmd is set, merge the metadata from this mmd to main_mmd,
    # otherwise our current mmd is main_mmd.
    if main_mmd:
        # Check for components that are in both MMDs before merging since MBS
        # currently can't handle that situation.
        duplicate_components = [rpm for rpm in main_mmd.get_rpm_components().keys()
                                if rpm in mmd.get_rpm_components()]
        if duplicate_components:
            error_msg = (
                'The included module "{0}" in "{1}" have the following '
                'conflicting components: {2}'.format(
                    mmd.get_name(), main_mmd.get_name(), ', '.join(duplicate_components)))
            raise UnprocessableEntity(error_msg)
        merge_included_mmd(main_mmd, mmd)
    else:
        main_mmd = mmd

    # If the modulemd yaml specifies components, then submit them for build
    rpm_components = mmd.get_rpm_components().values()
    module_components = mmd.get_module_components().values()
    all_components = list(rpm_components) + list(module_components)
    if not all_components:
        return

    rpm_weights = module_build_service.builder.GenericBuilder.get_build_weights(
        [c.get_name() for c in rpm_components])
    all_components.sort(key=lambda x: x.get_buildorder())
    # We do not start with batch = 0 here, because the first batch is
    # reserved for module-build-macros. First real components must be
    # planned for batch 2 and following.
    batch = initial_batch

    for component in all_components:
        # Increment the batch number when buildorder increases.
        if previous_buildorder != component.get_buildorder():
            previous_buildorder = component.get_buildorder()
            batch += 1

        # If the component is another module, we fetch its modulemd file
        # and record its components recursively with the initial_batch
        # set to our current batch, so the components of this module
        # are built in the right global order.
        if isinstance(component, Modulemd.ComponentModule):
            full_url = component.get_repository() + "?#" + component.get_ref()
            # It is OK to whitelist all URLs here, because the validity
            # of every URL have been already checked in format_mmd(...).
            included_mmd = _fetch_mmd(full_url, whitelist_url=True)[0]
            batch = record_component_builds(included_mmd, module, batch,
                                            previous_buildorder, main_mmd, session=session)
            continue

        component_ref = mmd.get_xmd()['mbs']['rpms'][component.get_name()]['ref']
        full_url = component.get_repository() + "?#" + component_ref
        build = models.ComponentBuild(
            module_id=module.id,
            package=component.get_name(),
            format="rpms",
            scmurl=full_url,
            batch=batch,
            ref=component_ref,
            weight=rpm_weights[component.get_name()]
        )
        session.add(build)

    return batch


def submit_module_build_from_yaml(username, handle, stream=None, skiptests=False,
                                  optional_params=None):
    yaml_file = handle.read()
    mmd = load_mmd(yaml_file)
    dt = datetime.utcfromtimestamp(int(time.time()))
    def_name = str(os.path.splitext(os.path.basename(handle.filename))[0])
    def_version = int(dt.strftime("%Y%m%d%H%M%S"))
    mmd.set_name(mmd.get_name() or def_name)
    mmd.set_stream(stream or mmd.get_stream() or "master")
    mmd.set_version(mmd.get_version() or def_version)
    if skiptests:
        buildopts = mmd.get_rpm_buildopts()
        buildopts["macros"] = buildopts.get("macros", "") + "\n\n%__spec_check_pre exit 0\n"
        mmd.set_rpm_buildopts(buildopts)
    return submit_module_build(username, None, mmd, optional_params)


_url_check_re = re.compile(r"^[^:/]+:.*$")


def submit_module_build_from_scm(username, url, branch, allow_local_url=False,
                                 optional_params=None):
    # Translate local paths into file:// URL
    if allow_local_url and not _url_check_re.match(url):
        log.info(
            "'{}' is not a valid URL, assuming local path".format(url))
        url = os.path.abspath(url)
        url = "file://" + url
    mmd, scm = _fetch_mmd(url, branch, allow_local_url)

    return submit_module_build(username, url, mmd, optional_params)


def _apply_dep_overrides(mmd, optional_params):
    """
    Apply the dependency override parameters (if specified) on the input modulemd.

    :param Modulemd.Module mmd: the modulemd to apply the overrides on
    :param dict optional_params: the optional API parameters passed in by the user
    :raises ValidationError: if one of the overrides doesn't apply
    """
    dep_overrides = {
        'buildrequires': optional_params.get('buildrequire_overrides', {}),
        'requires': optional_params.get('require_overrides', {})
    }

    unused_dep_overrides = {
        'buildrequires': set(dep_overrides['buildrequires'].keys()),
        'requires': set(dep_overrides['requires'].keys())
    }

    deps = mmd.get_dependencies()
    for dep in deps:
        for dep_type, overrides in dep_overrides.items():
            overridden = False
            # Get the existing streams (e.g. dep.get_buildrequires())
            reqs = getattr(dep, 'get_' + dep_type)()
            for name, streams in dep_overrides[dep_type].items():
                if name in reqs:
                    reqs[name].set(streams)
                    unused_dep_overrides[dep_type].remove(name)
                    overridden = True
            if overridden:
                # Set the overridden streams (e.g. dep.set_buildrequires(reqs))
                getattr(dep, 'set_' + dep_type)(reqs)

    for dep_type in unused_dep_overrides.keys():
        if unused_dep_overrides[dep_type]:
            raise ValidationError(
                'The {} overrides for the following modules aren\'t applicable: {}'
                .format(dep_type[:-1], ', '.join(sorted(unused_dep_overrides[dep_type]))))

    # Delete the parameters or else they'll get past to ModuleBuild.create
    if 'buildrequire_overrides' in optional_params:
        del optional_params['buildrequire_overrides']
    if 'require_overrides' in optional_params:
        del optional_params['require_overrides']

    mmd.set_dependencies(deps)


def submit_module_build(username, url, mmd, optional_params=None):
    """
    Submits new module build.

    :param str username: Username of the build's owner.
    :param str url: SCM URL of submitted build.
    :param Modulemd.Module mmd: Modulemd defining the build.
    :param dict optional_params: Dict with optional params for a build:
        - "local_build" (bool): The module is being built locally (the MBS is
          not running in infra, but on local developer's machine).
        - "default_streams" (dict): Dict with name:stream mapping defining the stream
          to choose for given module name if multiple streams are available to choose from.
        - Any optional ModuleBuild class field (str).
    :rtype: list with ModuleBuild
    :return: List with submitted module builds.
    """
    import koji  # Placed here to avoid py2/py3 conflicts...

    raise_if_stream_ambigous = False
    default_streams = {}
    if optional_params:
        # For local builds, we want the user to choose the exact stream using the default_streams
        # in case there are multiple streams to choose from and raise an exception otherwise.
        if "local_build" in optional_params:
            raise_if_stream_ambigous = True
            del optional_params["local_build"]
        # Get the default_streams if set.
        if "default_streams" in optional_params:
            default_streams = optional_params["default_streams"]
            del optional_params["default_streams"]
        _apply_dep_overrides(mmd, optional_params)

    validate_mmd(mmd)
    mmds = generate_expanded_mmds(db.session, mmd, raise_if_stream_ambigous, default_streams)
    modules = []

    for mmd in mmds:
        # Prefix the version of the modulemd based on the base module it buildrequires
        version = get_prefixed_version(mmd)
        mmd.set_version(version)
        version_str = str(version)

        log.debug('Checking whether module build already exists: %s.',
                  ":".join([mmd.get_name(), mmd.get_stream(), version_str, mmd.get_context()]))
        module = models.ModuleBuild.get_build_from_nsvc(
            db.session, mmd.get_name(), mmd.get_stream(), version_str, mmd.get_context())
        if module:
            if module.state != models.BUILD_STATES['failed']:
                err_msg = ('Module (state=%s) already exists. Only a new build or resubmission of '
                           'a failed build is allowed.' % module.state)
                log.error(err_msg)
                raise Conflict(err_msg)
            if optional_params:
                rebuild_strategy = optional_params.get('rebuild_strategy')
                if rebuild_strategy and module.rebuild_strategy != rebuild_strategy:
                    raise ValidationError(
                        'You cannot change the module\'s "rebuild_strategy" when '
                        'resuming a module build')
            log.debug('Resuming existing module build %r' % module)
            # Reset all component builds that didn't complete
            for component in module.component_builds:
                if component.state and component.state != koji.BUILD_STATES['COMPLETE']:
                    component.state = None
                    component.state_reason = None
                    db.session.add(component)
            module.username = username
            prev_state = module.previous_non_failed_state
            if prev_state == models.BUILD_STATES['init']:
                transition_to = models.BUILD_STATES['init']
            else:
                transition_to = models.BUILD_STATES['wait']
                module.batch = 0
            module.transition(conf, transition_to, "Resubmitted by %s" % username)
            log.info("Resumed existing module build in previous state %s" % module.state)
        else:
            log.info('Start to handle potential module stream collision.')
            record_stream_collision_modules(mmd)

            log.debug('Creating new module build')
            module = models.ModuleBuild.create(
                db.session,
                conf,
                name=mmd.get_name(),
                stream=mmd.get_stream(),
                version=version_str,
                modulemd=mmd.dumps(),
                scmurl=url,
                username=username,
                **(optional_params or {})
            )
            (module.ref_build_context, module.build_context, module.runtime_context,
             module.context) = module.contexts_from_mmd(module.modulemd)

        db.session.add(module)
        db.session.commit()
        modules.append(module)
        log.info("%s submitted build of %s, stream=%s, version=%s, context=%s", username,
                 mmd.get_name(), mmd.get_stream(), version_str, mmd.get_context())
    return modules


def _is_eol_in_pdc(name, stream):
    """ Check PDC if the module name:stream is no longer active. """

    params = {'type': 'module', 'global_component': name, 'name': stream}
    url = conf.pdc_url + '/component-branches/'

    response = requests.get(url, params=params)
    if not response:
        raise ValidationError("Failed to talk to PDC {}{}".format(response, response.text))

    data = response.json()
    results = data['results']
    if not results:
        raise ValidationError("No such module {}:{} found at {}".format(
            name, stream, response.request.url))

    # If the module is active, then it is not EOL and vice versa.
    return not results[0]['active']


def _fetch_mmd(url, branch=None, allow_local_url=False, whitelist_url=False,
               mandatory_checks=True):
    # Import it here, because SCM uses utils methods
    # and fails to import them because of dep-chain.
    import module_build_service.scm

    td = None
    scm = None
    try:
        log.debug('Verifying modulemd')
        td = tempfile.mkdtemp()
        if whitelist_url:
            scm = module_build_service.scm.SCM(url, branch, [url], allow_local_url)
        else:
            scm = module_build_service.scm.SCM(url, branch, conf.scmurls, allow_local_url)
        scm.checkout(td)
        if not whitelist_url and mandatory_checks:
            scm.verify()
        cofn = scm.get_module_yaml()
        mmd = load_mmd(cofn, is_file=True)
    finally:
        try:
            if td is not None:
                shutil.rmtree(td)
        except Exception as e:
            log.warning(
                "Failed to remove temporary directory {!r}: {}".format(
                    td, str(e)))

    if conf.check_for_eol:
        if _is_eol_in_pdc(scm.name, scm.branch):
            raise ValidationError(
                'Module {}:{} is marked as EOL in PDC.'.format(scm.name, scm.branch))

    if not mandatory_checks:
        return mmd, scm

    # If the name was set in the modulemd, make sure it matches what the scmurl
    # says it should be
    if mmd.get_name() and mmd.get_name() != scm.name:
        if not conf.allow_name_override_from_scm:
            raise ValidationError('The name "{0}" that is stored in the modulemd '
                                  'is not valid'.format(mmd.get_name()))
    else:
        mmd.set_name(scm.name)

    # If the stream was set in the modulemd, make sure it matches what the repo
    # branch is
    if mmd.get_stream() and mmd.get_stream() != scm.branch:
        if not conf.allow_stream_override_from_scm:
            raise ValidationError('The stream "{0}" that is stored in the modulemd '
                                  'does not match the branch "{1}"'.format(
                                      mmd.get_stream(), scm.branch))
    else:
        mmd.set_stream(str(scm.branch))

    # If the version is in the modulemd, throw an exception since the version
    # since the version is generated by MBS
    if mmd.get_version():
        raise ValidationError('The version "{0}" is already defined in the '
                              'modulemd but it shouldn\'t be since the version '
                              'is generated based on the commit time'.format(
                                  mmd.get_version()))
    else:
        mmd.set_version(int(scm.version))

    return mmd, scm


def load_mmd(yaml, is_file=False):
    try:
        if is_file:
            mmd = Modulemd.Module().new_from_file(yaml)
        else:
            mmd = Modulemd.Module().new_from_string(yaml)
        # If the modulemd was v1, it will be upgraded to v2
        mmd.upgrade()
    except Exception:
        error = 'The following invalid modulemd was encountered: {0}'.format(yaml)
        log.exception(error)
        raise UnprocessableEntity(error)

    return mmd


def load_local_builds(local_build_nsvs, session=None):
    """
    Loads previously finished local module builds from conf.mock_resultsdir
    and imports them to database.

    :param local_build_nsvs: List of NSV separated by ':' defining the modules
        to load from the mock_resultsdir.
    """
    if not local_build_nsvs:
        return

    if not session:
        session = db.session

    if type(local_build_nsvs) != list:
        local_build_nsvs = [local_build_nsvs]

    # Get the list of all available local module builds.
    builds = []
    try:
        for d in os.listdir(conf.mock_resultsdir):
            m = re.match('^module-(.*)-([^-]*)-([0-9]+)$', d)
            if m:
                builds.append((m.group(1), m.group(2), int(m.group(3)), d))
    except OSError:
        pass

    # Sort with the biggest version first
    try:
        # py27
        builds.sort(lambda a, b: -cmp(a[2], b[2]))  # noqa: F821
    except TypeError:
        # py3
        builds.sort(key=lambda a: a[2], reverse=True)

    for nsv in local_build_nsvs:
        parts = nsv.split(':')
        if len(parts) < 1 or len(parts) > 3:
            raise RuntimeError(
                'The local build "{0}" couldn\'t be be parsed into '
                'NAME[:STREAM[:VERSION]]'.format(nsv))

        name = parts[0]
        stream = parts[1] if len(parts) > 1 else None
        version = int(parts[2]) if len(parts) > 2 else None

        found_build = None
        for build in builds:
            if name != build[0]:
                continue
            if stream is not None and stream != build[1]:
                continue
            if version is not None and version != build[2]:
                continue

            found_build = build
            break

        if not found_build:
            raise RuntimeError(
                'The local build "{0}" couldn\'t be found in "{1}"'.format(
                    nsv, conf.mock_resultsdir))

        # Load the modulemd metadata.
        path = os.path.join(conf.mock_resultsdir, found_build[3], 'results')
        mmd = load_mmd(os.path.join(path, 'modules.yaml'), is_file=True)

        # Create ModuleBuild in database.
        module = models.ModuleBuild.create(
            session,
            conf,
            name=mmd.get_name(),
            stream=mmd.get_stream(),
            version=str(mmd.get_version()),
            context=mmd.get_context(),
            modulemd=mmd.dumps(),
            scmurl="",
            username="mbs",
            publish_msg=False)
        module.koji_tag = path
        module.state = models.BUILD_STATES['ready']
        session.commit()

        if (found_build[0] != module.name or found_build[1] != module.stream or
                str(found_build[2]) != module.version):
            raise RuntimeError(
                'Parsed metadata results for "{0}" don\'t match the directory name'
                .format(found_build[3]))
        log.info("Loaded local module build %r", module)



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

from flask.json import dumps
from flask import request

from module_build_service import app


def jsonify(*args, **kwargs):
    # This is `flask.jsonify` version which supports Python list as an input.
    # We cannot use real `jsonify`, because it can handle Python lists as
    # input only since 0.11, but RHEL7 contains 0.10.1.
    # https://github.com/pallets/flask/commit/daceb3e3a028b4b408c4bbdbdef0047f1de3a7c9
    indent = None
    separators = (',', ':')

    if app.config['JSONIFY_PRETTYPRINT_REGULAR'] and not request.is_xhr:
        indent = 2
        separators = (', ', ': ')

    if args and kwargs:
        raise TypeError("jsonify() behavior undefined when passed both args and kwargs")
    elif len(args) == 1:  # single args are passed directly to dumps()
        data = args[0]
    elif args:  # convert multiple args into an array
        data = list(args)
    else:  # convert kwargs to a dict
        data = dict(kwargs)

    # Note that we add '\n' to end of response
    # (see https://github.com/mitsuhiko/flask/pull/1262)
    rv = app.response_class(
        (dumps(data, indent=indent, separators=separators), '\n'),
        mimetype='application/json')
    return rv



# -*- coding: utf-8 -*-
#
# Copyright © 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluža <jkaluza@redhat.com>
#            Igor Gnatenko <ignatenko@redhat.com>

import enum
import collections
import itertools
import solv
from module_build_service import log, conf
from module_build_service.models import ModuleBuild


class MMDResolverPolicy(enum.Enum):
    All = "all"      # All possible top-level combinations
    First = "first"  # All possible top-level combinations (filtered by N:S, first picked)


class MMDResolver(object):
    """
    Resolves dependencies between Module metadata objects.
    """

    def __init__(self):
        self.pool = solv.Pool()
        self.pool.setarch("x86_64")
        self.build_repo = self.pool.add_repo("build")
        self.available_repo = self.pool.add_repo("available")
        # Solvable objects representing modules stored in a list grouped by
        # the name:stream.
        self.solvables = {}

    def _deps2reqs(self, deps, base_module_stream_overrides=None, exact_versions=True):
        """
        Helper method converting dependencies from MMD to solv.Dep instance expressing
        the dependencies in a way libsolv accepts as input.

        So for example for following input:
            deps = [{'gtk': ['1'], 'foo': ['1']}]
        The resulting solv.Dep expression will be:
            ((module(gtk) with module(gtk:1)) and (module(foo) with module(foo:1)))

        Base modules are handled in a special way in case when the stream of base module
        contains version in the "x.y.z" format. For example "el8.0.0" or "el7.6.0".
        In this case, the resulting solv.Dep expression for such base module will contain verison
        string computed using ModuleBuild.get_stream_version() method:
        For example:
            module(platform) with module(platform:el8) = 080200

        The stream used to compute the version can be also overriden using the
        `base_module_stream_overrides` dict which has base module name as a key and
        the stream which will be used to compute the version as a value.
        This is needed for cases when module requires just "platform:el8", but was
        in fact built against particular platform stream, for example platform:el8.1.0.
        In this case, such module should still require platform:el8, but in particular
        version which is passed to this method using the `base_module_stream_overrides`.

        When `exact_versions` is set to False, the base module dependency will contain
        ">=" operator instead of "=".

        The "with" syntax is here to allow depending on "module(gtk)" meaning "any gtk".
        This can happen in case {'gtk': []} is used as an input.

        See the inline comments for more information.

        :param list deps: List of dicts with dependency name as key and list of
            streams as value. Generally, it is just the return value from
            ``Modulemd.Dependencies.get_requires`` or
            ``Modulemd.Dependencies.get_buildrequires`` whose value is
            converted from ``Modulemd.SimpleSet`` to list.
        :param dict base_module_stream_overrides: The key is base module name, value
            is the stream string which will be used to compute `version` part of the
            base module solv.Dep expression.
        :param bool exact_versions: When set to False, the base module dependency
            will contain ">=" operator instead of "=".
        :rtype: solv.Dep
        :return: solv.Dep instance with dependencies in form libsolv accepts.
        """
        pool = self.pool

        # Every name:stream combination from dict in `deps` list is expressed as `solv.Dep`
        # instance and is represented internally in solv with "module(name:stream)".
        # This is parallel to RPM-world "Provides: perl(foo)" or "Requires: perl(foo)",
        # but in this method, we are only constructing the condition after the "Provides:"
        # or "Requires:".
        # This method creates such solve.Dep.
        stream_dep = lambda n, s: pool.Dep("module(%s:%s)" % (n, s))
        versioned_stream_dep = lambda n, s, v, op: pool.Dep("module(%s:%s)" % (n, s)).Rel(
            op, pool.Dep(str(v)))

        # There are relations between modules in `deps`. For example:
        #   deps = [{'gtk': ['1'], 'foo': ['1']}]" means "gtk:1 and foo:1" are both required.
        #   deps = [{'gtk': ['1', '2']}"] means "gtk:1 or gtk:2" are required.
        # This method helps creating such relations using following syntax:
        #   rel_or_dep(solv.Dep, solve.REL_OR, stream_dep(name, stream))
        #   rel_or_dep(solv.Dep, solve.REL_AND, stream_dep(name, stream))
        #   rel_or_dep(solv.Dep, solve.REL_WITH, stream_dep(name, stream))
        #   rel_or_dep(solv.Dep, solve.REL_WITHOUT, stream_dep(name, stream))
        rel_or_dep = lambda dep, op, rel: dep.Rel(op, rel) if dep is not None else rel

        # Check each dependency dict in `deps` list and generate the solv requirements.
        reqs = None
        for dep_dicts in deps:
            # Contains the solv.Dep requirements for current dict.
            require = None
            for name, streams in dep_dicts.items():
                is_base_module = name in conf.base_module_names

                # The req_pos will store solv.Dep expression for "positive" requirements.
                # That is the case of 'gtk': ['1', '2'].
                # The req_neg will store negative requirements like 'gtk': ['-1', '-2'].
                req_pos = req_neg = None

                # For each stream in `streams` for this dependency, generate the
                # module(name:stream) solv.Dep and add REL_OR relations between them.
                for stream in streams:
                    if is_base_module:
                        # Override the stream which is used to compute the stream version in case
                        # `base_module_stream_overrides` is set.
                        if base_module_stream_overrides and name in base_module_stream_overrides:
                            stream_for_version = base_module_stream_overrides[name]
                        else:
                            stream_for_version = stream

                        # In case x.y.z versioning is not used for this base module, do not
                        # use versions solv.Dep.
                        if len(str(ModuleBuild.get_stream_version(
                                stream_for_version, right_pad=False))) < 5:
                            if stream.startswith("-"):
                                req_neg = rel_or_dep(
                                    req_neg, solv.REL_OR, stream_dep(name, stream[1:]))
                            else:
                                req_pos = rel_or_dep(
                                    req_pos, solv.REL_OR, stream_dep(name, stream))
                        else:
                            # The main reason why to use `exact_versions` is the case when
                            # adding deps for the input module we want to resolve. This module
                            # buildrequires exact stream version of base module against which it
                            # needs for building and we should never pull in different one.
                            # But for modules which are buildrequires of this input module, we
                            # want to use "base_module >= stream_version" relation, because they
                            # can be chery-picked even when newer base module stream_version is
                            # requested, for example:
                            # - foo buildrequires bar and also buildrequires platform:el8 = 080100.
                            # - bar:1 is built against platform:el8.0.0.
                            # - bar:2 is built against platform:el8.2.0.
                            # We need libsolv to allow chery-picking "bar:1", and ignore "bar:2",
                            # because it is built against newer platform stream version than the
                            # requested and and such newer version can be incompatible with the
                            # old one. so we express bar's dependencies on platform like this:
                            # - bar:1 buildrequires platform:el8 >= 080000.
                            # - bar:2 buildrequires platform:el8 >= 080200.
                            # Because the "foo" limits the solving to platform:el8 = 080100,
                            # the bar:2 won't be returned by libsolv, because 080100 < 080200.
                            # But that bar:1 will be returned by libsovl, because it buildrequires
                            # platform 080000 which is lesser than 080100.
                            op = solv.REL_EQ
                            if not exact_versions:
                                op |= solv.REL_GT
                            version = ModuleBuild.get_stream_version(
                                stream_for_version, right_pad=False)
                            if stream.startswith("-"):
                                req_neg = rel_or_dep(
                                    req_neg, solv.REL_OR,
                                    versioned_stream_dep(name, stream[1:], version, op))
                            else:
                                req_pos = rel_or_dep(
                                    req_pos, solv.REL_OR,
                                    versioned_stream_dep(name, stream, version, op))
                    else:
                        if stream.startswith("-"):
                            req_neg = rel_or_dep(
                                req_neg, solv.REL_OR, stream_dep(name, stream[1:]))
                        else:
                            req_pos = rel_or_dep(req_pos, solv.REL_OR, stream_dep(name, stream))

                # Generate the module(name) solv.Dep.
                req = pool.Dep("module(%s)" % name)

                # Use the REL_WITH for positive requirements and REL_WITHOUT for negative
                # requirements.
                if req_pos is not None:
                    req = req.Rel(solv.REL_WITH, req_pos)
                elif req_neg is not None:
                    req = req.Rel(solv.REL_WITHOUT, req_neg)

                # And in the end use AND between the last name:[streams] and the current one.
                require = rel_or_dep(require, solv.REL_AND, req)

            # There might be multiple dicts in `deps` list, so use OR relation between them.
            reqs = rel_or_dep(reqs, solv.REL_OR, require)

        return reqs

    def _add_virtual_provides(self, solvable, mmd):
        """
        Adds "virtual_streams" from XMD section of `mmd` to `solvable`.

        Base modules like "platform" can contain virtual streams which need to be considered
        when resolving dependencies. For example module "platform:el8.1.0" can provide virtual
        stream "el8". In this case the solvable will have following additional Provides:

        - module(platform:el8.1.0) = 80100 - Modules can require specific platform stream.
        - module(platform:el8) = 80100 - Module can also require just platform:el8.
        """
        xmd = mmd.get_xmd()
        # Return in case virtual_streams are not set for this mmd.
        if "mbs" not in xmd.keys() or "virtual_streams" not in xmd["mbs"].keys():
            return

        # When depsolving, we will need to follow specific rules to choose the right base
        # module, like sorting the base modules sharing the same virtual streams based on
        # their "stream version" - For example stream "el8.1" is lower than stream "el8.2"
        # and so on. We therefore need to convert the stream and version of base module to
        # integer representation and add "module($name:$stream) = $stream_based_version"
        # to Provides.
        version = ModuleBuild.get_stream_version(
            mmd.get_stream(), right_pad=False)
        dep = self.pool.Dep("module(%s:%s)" % (mmd.get_name(), mmd.get_stream())).Rel(
            solv.REL_EQ, self.pool.Dep(str(version)))
        solvable.add_deparray(solv.SOLVABLE_PROVIDES, dep)

        # For each virtual stream, add
        # "module($name:$stream) = $virtual_stream_based_version" provide.
        for stream in xmd["mbs"]["virtual_streams"]:
            dep = self.pool.Dep("module(%s:%s)" % (mmd.get_name(), stream)).Rel(
                solv.REL_EQ, self.pool.Dep(str(version)))
            solvable.add_deparray(solv.SOLVABLE_PROVIDES, dep)

    def _get_base_module_stream_overrides(self, mmd):
        """
        Checks the xmd["mbs"]["buildrequires"] and returns the dict containing
        base module name as a key and stream of base module against which this
        module was built. This is later used to override base module streams
        in the _deps2reqs method.

        :param Modulemd mmd: Metadata of module for which the stream overrides are returned.
        :rtype: dict
        :return: Dict with module name as a key and new stream as a value.
        """
        overrides = {}
        if "mbs" in mmd.get_xmd().keys():
            for base_module_name in conf.base_module_names:
                if base_module_name not in mmd.get_xmd()["mbs"]["buildrequires"].keys():
                    continue
                if "stream" not in mmd.get_xmd()["mbs"]["buildrequires"][base_module_name].keys():
                    continue
                stream = mmd.get_xmd()["mbs"]["buildrequires"][base_module_name]["stream"]

                overrides[base_module_name] = stream
        return overrides

    def add_modules(self, mmd):
        """
        Adds module represented by `mmd` metadata to MMDResolver. Modules added by this
        method will be considered as possible dependencies while resolving the dependencies
        using the `solve(...)` method only if their "context" is None. Otherwise they are
        treated like input modules we want to resolve dependencies for.

        :param Modulemd mmd: Metadata of module to add.
        :rtype: list
        :return: list of solv.Solvable instances representing the module in libsolv world.
        """
        n, s, v, c = mmd.get_name(), mmd.get_stream(), mmd.get_version(), mmd.get_context()
        pool = self.pool

        # Helper method to return the dependencies of `mmd` in the {name: [streams], ... form}.
        # The `fn` is either "get_requires" or "get_buildrequires" str depending on whether
        # the return deps should be runtime requires or buildrequires.
        normdeps = lambda mmd, fn: [{name: streams.get()
                                     for name, streams in getattr(dep, fn)().items()}
                                    for dep in mmd.get_dependencies()]

        base_module_stream_overrides = self._get_base_module_stream_overrides(mmd)

        # Each solvable object has name, version, architecture and list of
        # provides/requires/conflicts which defines its relations with other solvables.
        # You can imagine solvable as a single RPM.
        # Single module can be represented by multiple solvables - read further inline
        # comments for more info. Therefore we use list to store them.
        solvables = []
        if c is not None:
            # If context is set, the module we are adding should be used as dependencies
            # for input module. Therefore add it in "available_repo".
            solvable = self.available_repo.add_solvable()

            # Use n:s:v:c as name, version as version and set the arches.
            solvable.name = "%s:%s:%d:%s" % (n, s, v, c)
            solvable.evr = str(v)
            # TODO: replace with real arch, but for now resolving using single arch
            # is sufficient.
            solvable.arch = "x86_64"

            # Add "Provides: module(name)", each module provides itself.
            # This is used for example to find the buildrequired module when
            # no particular stream is used - for example when buildrequiring
            # "gtk: []"
            solvable.add_deparray(solv.SOLVABLE_PROVIDES,
                                  pool.Dep("module(%s)" % n))
            # Add "Provides: module(name:stream) = version", so we can find buildrequired
            # modules when "gtk:[1]" is used and also choose the latest version.
            solvable.add_deparray(solv.SOLVABLE_PROVIDES,
                                  pool.Dep("module(%s:%s)" % (n, s)).Rel(
                                      solv.REL_EQ, pool.Dep(str(v))))

            self._add_virtual_provides(solvable, mmd)

            # Fill in the "Requires" of this module, so we can track its dependencies
            # on other modules.
            requires = self._deps2reqs(normdeps(mmd, "get_requires"),
                                       base_module_stream_overrides, False)
            log.debug("Adding module %s with requires: %r", solvable.name, requires)
            solvable.add_deparray(solv.SOLVABLE_REQUIRES, requires)

            # Add "Conflicts: module(name)".
            # This is needed to prevent installation of multiple streams of single module.
            # For example:
            #  - "app:1" requires "foo:1" and "gtk:1".
            #  - "foo:1" requires "bar:1".
            #  - "gtk:1" requires "bar:2".
            # "bar:1" and "bar:2" cannot be installed in the same time and therefore
            # there need to be conflict defined between them.
            solvable.add_deparray(solv.SOLVABLE_CONFLICTS, pool.Dep("module(%s)" % n))
            solvables.append(solvable)

            # Add solvable to solvables list. Sorting is done later in the solve method.
            ns = ":".join([n, s])
            if ns not in self.solvables:
                self.solvables[ns] = []
            self.solvables[ns].append(solvable)
        else:
            # For input module, we might have multiple buildrequires/requires pairs in the
            # input `mmd`. For example like this:
            #   - buildrequires:
            #       gtk: [1]
            #       platform: [f28]
            #     requires:
            #       gtk: [1]
            #   - buildrequires:
            #       gtk: [2]
            #       platform: [f29]
            #     requires:
            #       gtk: [2]
            # This means we need: "(gtk:1 and platform:f28) or (gtk:2 and platform:f29)".
            # There is no way how to express that in libsolv as single solvable and in the same
            # time try all the possible combinations. Libsolv just returns the single one and does
            # not offer enough data for us to tell it to try another one to really find all of
            # them.
            # The solution for that is therefore adding multiple solvables for each OR block of
            # that input condition.
            #
            # So in our example, we add two solvables:
            #   1) Solvable with name "n:s:v:0" and "Requires: gtk:1 and platform:f28".
            #   2) Solvable with name "n:s:v:1" and "Requires: gtk:2 and platform:f29".
            #
            # Note the "context" field in the solvable name - it is set according to index
            # of buildrequires/requires pair and uniquely identifies the Solvable.
            #
            # Using this trick, libsolv will try to solve all the buildrequires/requires pairs,
            # because they are expressed as separate Solvables and we are able to distinguish
            # between them thanks to context value.
            normalized_deps = normdeps(mmd, "get_buildrequires")
            for c, deps in enumerate(mmd.get_dependencies()):
                # $n:$s:$c-$v.src
                solvable = self.build_repo.add_solvable()
                solvable.name = "%s:%s:%d:%d" % (n, s, v, c)
                solvable.evr = str(v)
                solvable.arch = "src"

                requires = self._deps2reqs([normalized_deps[c]], base_module_stream_overrides)
                log.debug("Adding module %s with requires: %r", solvable.name, requires)
                solvable.add_deparray(solv.SOLVABLE_REQUIRES, requires)

                solvables.append(solvable)
                # Add solvable to solvables list. Sorting is done later in the solve method.
                ns = ":".join([n, s])
                if ns not in self.solvables:
                    self.solvables[ns] = []
                self.solvables[ns].append(solvable)

        return solvables

    def solve(self, mmd, policy=MMDResolverPolicy.First):
        """
        Solves dependencies of module defined by `mmd` object. Returns set
        containing frozensets with all the possible combinations which
        satisfied dependencies.

        :param Modulemd mmd: Input modulemd which should have the `context` set
            to None.
        :param policy: Policy to use when the dependencies used in buildrequires
            section are ambigous. For example, when the single buildrequired
            module is gtk:1 and this gtk:1 module is built against both
            platform:f28 and platform:f29, the policy influences the resolving
            in following way:

            - MMDResolverPolicy.First: Only single combination of buildrequires
              will be returned with "gtk:1" and "platform:f28", because the input
              buildrequires section did not mention any platform stream and
              therefore "first one" is used.
            - MMDResolverPolicy.All: Two combinations of buildrequires will be returned,
              one with "gtk:1" and "platform:f28", other with "gtk:1" and "platform:f29".

        :return: set of frozensets of n:s:v:c of modules which satisfied the
            dependency solving.
        """
        # Add the input module to pool and generate the "Provides" data so we can
        # use them for resolving later.
        solvables = self.add_modules(mmd)
        if not solvables:
            raise ValueError("No module(s) found for resolving")
        self.pool.createwhatprovides()

        # "solvable to n:s:v:c"
        s2nsvc = lambda s: "%s:%s" % (s.name, s.arch)
        # "solvable to n:s"
        s2ns = lambda s: ":".join(s.name.split(":", 2)[:2])

        # Sort the solvables by version for NSVC in descending order.
        for ns, unordered_solvables in self.solvables.items():
            unordered_solvables.sort(key=lambda s: int(s.name.split(":")[2]), reverse=True)

        # For each solvable object generated from input module, run the solver.
        # For reasons why there might be multiple solvable objects, please read the
        # `add_modules(...)` inline comments.
        solver = self.pool.Solver()
        alternatives = collections.OrderedDict()
        for src in solvables:
            # Create the solv Job to represent the solving task.
            job = self.pool.Job(solv.Job.SOLVER_INSTALL | solv.Job.SOLVER_SOLVABLE, src.id)

            # Check that at max 1 requires element exists in the solvable object - since
            # we create multiple solvable objects where each of them has at max one
            # requires element, it should never be the case...
            # NOTE: "requires" in solvable are actually "buildrequires" in mmd.
            requires = src.lookup_deparray(solv.SOLVABLE_REQUIRES)
            if len(requires) > 1:
                raise SystemError("At max one element should be in Requires: %s" % requires)
            elif len(requires) == 0:
                # Return early in case the requires is empty, because it basically means
                # the module has no buildrequires section.
                return set([frozenset([s2nsvc(src)])])

            requires = requires[0]
            src_alternatives = alternatives[src] = collections.OrderedDict()

            # TODO: replace this ugliest workaround ever with sane code of parsing rich deps.
            # We need to split them because whatprovides() treats "and" same way as "or" which is
            # not enough to generate combinations.
            # Source solvables have Req: (X and Y and Z)
            # Binary solvables have Req: ((X and Y) or (X and Z))
            # They do use "or" within "and", so simple string split won't work for binary packages.
            if src.arch != "src":
                raise NotImplementedError

            # What we get in `requires` here is a string in following format:
            #   ((module(gtk) with module(gtk:1)) and (module(foo) with module(foo:1)) and (...))
            # And what we want to get is the list of all valid combinations with particular NSVCs
            # of buildrequired modules. There are few steps we need to do to achieve that:

            # 1) Convert the "(R1 and R2 and R3)" string to list of solv.Dep in following format:
            #    [solv.Dep(R1), solv.Dep(R2), solv.Dep(R3), ...]
            deps = str(requires).split(" and ")
            if len(deps) > 1:
                # Remove the extra parenthesis in the input string in case there are more
                # rules.
                deps[0] = deps[0][1:]
                deps[-1] = deps[-1][:-1]
            # Generate the new deps using the parserpmrichdep.
            deps = [self.pool.parserpmrichdep(dep) if dep.startswith("(") else self.pool.Dep(dep)
                    for dep in deps]

            # 2) For each dep (name:stream), get the set of all solvables in particular NSVCs,
            #    which provides that name:stream. Then use itertools.product() to actually
            #    generate all the possible combinations so we can try solving them.
            for opt in itertools.product(*[self.pool.whatprovides(dep) for dep in deps]):
                log.debug("Testing %s with combination: %s", src, opt)
                # We will be trying to solve all the combinations using all the NSVCs
                # we have in pool, but as we said earlier, we don't want to return
                # all of them when MMDResolverPolicy.First is used.
                # We will achieve that by storing alternative combinations in `src_alternatives`
                # with NSVC as key in case we want all of them and NS as a key when we want
                # just First combination for given dependency.
                # This will allow us to group alternatives for single NS in case of First
                # policy and later return just the first alternative.
                if policy == MMDResolverPolicy.All:
                    kfunc = s2nsvc
                elif policy == MMDResolverPolicy.First:
                    kfunc = s2ns
                key = tuple(kfunc(s) for s in opt)
                # `key` now contains tuple similar to "('gtk:1', 'foo:1')"
                alternative = src_alternatives.setdefault(key, [])

                # Create the solving jobs.
                # We need to say to libsolv that we want it to prefer modules from the combination
                # we are currently trying, otherwise it would just choose some random ones.
                # We do that by FAVORING those modules - this is done in libsolv by another
                # job prepending to our main job to resolve the deps of input module.
                jobs = [self.pool.Job(solv.Job.SOLVER_FAVOR | solv.Job.SOLVER_SOLVABLE, s.id)
                        for s in opt] + [job]

                # Log the job.
                log.debug("Jobs:")
                for j in jobs:
                    log.debug("  - %s", j)
                # Solve the deps and log the dependency issues.
                problems = solver.solve(jobs)
                if problems:
                    problem_str = self._detect_transitive_stream_collision(problems)
                    if problem_str:
                        err_msg = problem_str
                    else:
                        err_msg = ', '.join(str(p) for p in problems)
                    raise RuntimeError(
                        'Problems were found during module dependency resolution: {}'
                        .format(err_msg))
                # Find out what was actually resolved by libsolv to be installed as a result
                # of our jobs - those are the modules we are looking for.
                newsolvables = solver.transaction().newsolvables()
                log.debug("Transaction:")

                for s in newsolvables:
                    log.debug("  - %s", s)

                # Skip this alternative in case not all the favored Solvables are in
                # the solution. For example if we favored gtk:4:1:c8, but it simply
                # cannot be installed because of other dependencies, we know this is not a
                # possible combination and we should not treat it as an alternative.
                all_solvables_found = True
                for s in opt:
                    if s not in newsolvables:
                        all_solvables_found = False
                        break

                # Append them as an alternative for this src_alternative.
                # Remember that src_alternatives are grouped by NS or NSVC depending on
                # MMDResolverPolicy, so there might be more of them.
                if all_solvables_found:
                    alternative.append(newsolvables)
                else:
                    log.debug("  - ^ Not all favored solvables found in the result, skipping.")

        # If the MMDResolverPolicy is First, we will check all the alternatives and keep
        # just the "first" one.
        if policy == MMDResolverPolicy.First:
            # Prune
            for transactions in alternatives.values():
                for ns, trans in transactions.items():
                    # Each transaction in trans lists all the possible working
                    # combination of solvables. Our goal here is to find out the
                    # transaction which installs the most latest Solvables - ideally
                    # always the latest versions of the Solvables we have, but this might
                    # not be always possible because of dependencies.
                    #
                    # We achieve that by generating sorted_trans list in follwing format:
                    #   [[transaction_id, [solvable1_index, solvable2_index, ...]], [...], ...]
                    #
                    # The solvableN_index is a number saying how new the solvable is. We use
                    # `self.solvables` to get that number and it is simply index
                    # of the solvable in the particular self.solvables[name_stream] list.
                    # The newest solvable has therefore index 0, the next newest solvable index 1
                    # and so on.
                    #
                    # Then we simply sort the `sorted_trans` based on the sum of solvableN_index
                    # which gives us the transaction with the most recent versions. This is
                    # used as a solution.
                    sorted_trans = []
                    for i, t in enumerate(trans):
                        idx = []
                        for s in t:
                            name_stream = s2ns(s)
                            if name_stream not in self.solvables:
                                continue
                            index = self.solvables[name_stream].index(s)
                            idx.append(index)
                        sorted_trans.append([i, idx])
                    sorted_trans.sort(key=lambda i: sum(i[1]))
                    if sorted_trans:
                        transactions[ns] = [trans[sorted_trans[0][0]]]

        # Convert the solvables in alternatives to nsvc and return them as set of frozensets.
        return set(frozenset(s2nsvc(s) for s in transactions[0])
                   for src_alternatives in alternatives.values()
                   for transactions in src_alternatives.values())

    @staticmethod
    def _detect_transitive_stream_collision(problems):
        """Return problem description if transitive stream collision happens

        Transitive stream collision could happen if different buildrequired
        modules requires same module but with different streams. For example,

        app:1 --br--> gtk:1 --req--> baz:1* -------- req --------> platform:f29
             |                                                     ^
             +--br--> foo:1 --req--> bar:1 --req--> baz:2* --req---|

        as a result, ``baz:1`` will conflicts with ``baz:2``.

        :param problems: list of problems returned from ``solv.Solver.solve``.
        :return: a string of problem description if transitive stream collision
            is detected. The description is provided by libsolv without
            changed. If no such collision, None is returned.
        :rtype: str or None
        """

        def find_conflicts_pairs():
            for problem in problems:
                for rule in problem.findallproblemrules():
                    info = rule.info()
                    if info.type == solv.Solver.SOLVER_RULE_PKG_CONFLICTS:
                        pair = [info.solvable.name, info.othersolvable.name]
                        pair.sort()  # only for pretty print
                        yield pair

        formatted_conflicts_pairs = ', '.join(
            '{} and {}'.format(*item) for item in find_conflicts_pairs()
        )
        if formatted_conflicts_pairs:
            return 'The module has conflicting buildrequires of: {}'.format(
                formatted_conflicts_pairs)



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import koji
import os
from os import path, mkdir
from os.path import dirname
from shutil import copyfile
from datetime import datetime, timedelta
from random import randint

import module_build_service.messaging
import module_build_service.scheduler.handlers.repos
import module_build_service.utils
from module_build_service.errors import Forbidden
from module_build_service import db, models, conf, build_logs

from mock import patch, PropertyMock, Mock
from werkzeug.datastructures import FileStorage
import kobo
import pytest

from tests import app, reuse_component_init_data, clean_database
import json
import itertools

from module_build_service.builder.base import GenericBuilder
from module_build_service.builder.KojiModuleBuilder import KojiModuleBuilder
import module_build_service.scheduler.consumer
from module_build_service.messaging import MBSModule

base_dir = dirname(dirname(__file__))

user = ('Homer J. Simpson', set(['packager']))


class FakeSCM(object):
    def __init__(self, mocked_scm, name, mmd_filename, commit=None, version=20180205135154):
        self.mocked_scm = mocked_scm
        self.name = name
        self.commit = commit
        self.version = version
        self.mmd_filename = mmd_filename
        self.sourcedir = None

        self.mocked_scm.return_value.checkout = self.checkout
        self.mocked_scm.return_value.name = self.name
        self.mocked_scm.return_value.branch = 'master'
        self.mocked_scm.return_value.get_latest = self.get_latest
        self.mocked_scm.return_value.commit = self.commit
        self.mocked_scm.return_value.version = self.version
        self.mocked_scm.return_value.repository_root = "git://pkgs.stg.fedoraproject.org/modules/"
        self.mocked_scm.return_value.sourcedir = self.sourcedir
        self.mocked_scm.return_value.get_module_yaml = self.get_module_yaml

    def checkout(self, temp_dir):
        self.sourcedir = path.join(temp_dir, self.name)
        mkdir(self.sourcedir)
        base_dir = path.abspath(path.dirname(__file__))
        copyfile(path.join(base_dir, '..', 'staged_data', self.mmd_filename),
                 self.get_module_yaml())

        return self.sourcedir

    def get_latest(self, ref='master'):
        return ref

    def get_module_yaml(self):
        return path.join(self.sourcedir, self.name + ".yaml")


class FakeModuleBuilder(GenericBuilder):
    """
    Fake module builder which succeeds for every build.
    """

    backend = "test"
    BUILD_STATE = "COMPLETE"
    # Simulates a situation when a component is already built in Koji
    INSTANT_COMPLETE = False
    DEFAULT_GROUPS = None

    on_build_cb = None
    on_cancel_cb = None
    on_buildroot_add_artifacts_cb = None
    on_tag_artifacts_cb = None

    @module_build_service.utils.validate_koji_tag('tag_name')
    def __init__(self, owner, module, config, tag_name, components):
        self.module_str = module
        self.tag_name = tag_name
        self.config = config

    @classmethod
    def reset(cls):
        FakeModuleBuilder.BUILD_STATE = "COMPLETE"
        FakeModuleBuilder.INSTANT_COMPLETE = False
        FakeModuleBuilder.on_build_cb = None
        FakeModuleBuilder.on_cancel_cb = None
        FakeModuleBuilder.on_buildroot_add_artifacts_cb = None
        FakeModuleBuilder.on_tag_artifacts_cb = None
        FakeModuleBuilder.DEFAULT_GROUPS = None
        FakeModuleBuilder.backend = 'test'

    def buildroot_connect(self, groups):
        default_groups = FakeModuleBuilder.DEFAULT_GROUPS or {
            'srpm-build':
                set(['shadow-utils', 'fedora-release', 'redhat-rpm-config',
                     'rpm-build', 'fedpkg-minimal', 'gnupg2', 'bash']),
            'build':
                set(['unzip', 'fedora-release', 'tar', 'cpio', 'gawk',
                     'gcc', 'xz', 'sed', 'findutils', 'util-linux', 'bash',
                     'info', 'bzip2', 'grep', 'redhat-rpm-config',
                     'diffutils', 'make', 'patch', 'shadow-utils',
                     'coreutils', 'which', 'rpm-build', 'gzip', 'gcc-c++'])}
        if groups != default_groups:
            raise ValueError("Wrong groups in FakeModuleBuilder.buildroot_connect()")

    def buildroot_prep(self):
        pass

    def buildroot_resume(self):
        pass

    def buildroot_ready(self, artifacts=None):
        return True

    def buildroot_add_dependency(self, dependencies):
        pass

    def repo_from_tag(self, config, tag_name, arch):
        pass

    def buildroot_add_artifacts(self, artifacts, install=False):
        if FakeModuleBuilder.on_buildroot_add_artifacts_cb:
            FakeModuleBuilder.on_buildroot_add_artifacts_cb(self, artifacts, install)
        if self.backend == 'test':
            for nvr in artifacts:
                # buildroot_add_artifacts received a list of NVRs, but the tag message expects the
                # component name. At this point, the NVR may not be set if we are trying to reuse
                # all components, so we can't search the database. We must parse the package name
                # from the nvr and then tag it in the build tag. Kobo doesn't work when parsing
                # the NVR of a component with a module dist-tag, so we must manually do it.
                package_name = nvr.split('.module')[0].rsplit('-', 2)[0]
                # When INSTANT_COMPLETE is on, the components are already in the build tag
                if self.INSTANT_COMPLETE is False:
                    self._send_tag(package_name, nvr, dest_tag=False)
        elif self.backend == 'testlocal':
            self._send_repo_done()

    def buildroot_add_repos(self, dependencies):
        pass

    def tag_artifacts(self, artifacts, dest_tag=True):
        if FakeModuleBuilder.on_tag_artifacts_cb:
            FakeModuleBuilder.on_tag_artifacts_cb(self, artifacts, dest_tag=dest_tag)

        if self.backend == 'test':
            for nvr in artifacts:
                # tag_artifacts received a list of NVRs, but the tag message expects the
                # component name
                artifact = models.ComponentBuild.query.filter_by(nvr=nvr).first().package
                self._send_tag(artifact, nvr, dest_tag=dest_tag)

    @property
    def koji_session(self):
        session = Mock()

        def _newRepo(tag):
            session.newRepo = self._send_repo_done()
            return 123
        session.newRepo = _newRepo
        return session

    @property
    def module_build_tag(self):
        return {"name": self.tag_name + "-build"}

    def _send_repo_done(self):
        msg = module_build_service.messaging.KojiRepoChange(
            msg_id='a faked internal message',
            repo_tag=self.tag_name + "-build",
        )
        module_build_service.scheduler.consumer.work_queue_put(msg)

    def _send_tag(self, artifact, nvr, dest_tag=True):
        if dest_tag:
            tag = self.tag_name
        else:
            tag = self.tag_name + "-build"
        msg = module_build_service.messaging.KojiTagChange(
            msg_id='a faked internal message',
            tag=tag,
            artifact=artifact,
            nvr=nvr
        )
        module_build_service.scheduler.consumer.work_queue_put(msg)

    def _send_build_change(self, state, name, build_id):
        # build_id=1 and task_id=1 are OK here, because we are building just
        # one RPM at the time.
        msg = module_build_service.messaging.KojiBuildChange(
            msg_id='a faked internal message',
            build_id=build_id,
            task_id=build_id,
            build_name=name,
            build_new_state=state,
            build_release="1",
            build_version="1"
        )
        module_build_service.scheduler.consumer.work_queue_put(msg)

    def build(self, artifact_name, source):
        print("Starting building artifact %s: %s" % (artifact_name, source))
        build_id = randint(1000, 9999999)

        if FakeModuleBuilder.on_build_cb:
            FakeModuleBuilder.on_build_cb(self, artifact_name, source)

        if FakeModuleBuilder.BUILD_STATE != "BUILDING":
            self._send_build_change(
                koji.BUILD_STATES[FakeModuleBuilder.BUILD_STATE], artifact_name, build_id)

        reason = "Submitted %s to Koji" % (artifact_name)
        return build_id, koji.BUILD_STATES['BUILDING'], reason, None

    @staticmethod
    def get_disttag_srpm(disttag, module_build):
        # @FIXME
        return KojiModuleBuilder.get_disttag_srpm(disttag, module_build)

    def cancel_build(self, task_id):
        if FakeModuleBuilder.on_cancel_cb:
            FakeModuleBuilder.on_cancel_cb(self, task_id)

    def list_tasks_for_components(self, component_builds=None, state='active'):
        pass

    def recover_orphaned_artifact(self, component_build):
        msgs = []
        if self.INSTANT_COMPLETE:
            disttag = module_build_service.utils.get_rpm_release(
                component_build.module_build)
            # We don't know the version or release, so just use a random one here
            nvr = '{0}-1.0-1.{1}'.format(component_build.package, disttag)
            component_build.state = koji.BUILD_STATES['COMPLETE']
            component_build.nvr = nvr
            component_build.task_id = component_build.id + 51234
            component_build.state_reason = 'Found existing build'
            nvr_dict = kobo.rpmlib.parse_nvr(component_build.nvr)
            # Send a message stating the build is complete
            msgs.append(module_build_service.messaging.KojiBuildChange(
                'recover_orphaned_artifact: fake message', randint(1, 9999999),
                component_build.task_id, koji.BUILD_STATES['COMPLETE'], component_build.package,
                nvr_dict['version'], nvr_dict['release'], component_build.module_build.id))
            # Send a message stating that the build was tagged in the build tag
            msgs.append(module_build_service.messaging.KojiTagChange(
                'recover_orphaned_artifact: fake message',
                component_build.module_build.koji_tag + '-build', component_build.package,
                component_build.nvr))
        return msgs

    def finalize(self):
        pass


def cleanup_moksha():
    # Necessary to restart the twisted reactor for the next test.
    import sys
    del sys.modules['twisted.internet.reactor']
    del sys.modules['moksha.hub.reactor']
    del sys.modules['moksha.hub']
    import moksha.hub.reactor # noqa


@patch('module_build_service.utils.submit.record_stream_collision_modules')
@patch.object(module_build_service.config.Config, 'system', new_callable=PropertyMock,
              return_value='test')
@patch("module_build_service.builder.GenericBuilder.default_buildroot_groups",
       return_value={
           'srpm-build':
           set(['shadow-utils', 'fedora-release', 'redhat-rpm-config',
                'rpm-build', 'fedpkg-minimal', 'gnupg2', 'bash']),
           'build':
           set(['unzip', 'fedora-release', 'tar', 'cpio', 'gawk',
                'gcc', 'xz', 'sed', 'findutils', 'util-linux', 'bash',
                'info', 'bzip2', 'grep', 'redhat-rpm-config',
                'diffutils', 'make', 'patch', 'shadow-utils',
                'coreutils', 'which', 'rpm-build', 'gzip', 'gcc-c++'])})
class TestBuild:
    # Global variable used for tests if needed
    _global_var = None

    def setup_method(self, test_method):
        GenericBuilder.register_backend_class(FakeModuleBuilder)
        self.client = app.test_client()
        clean_database()

    def teardown_method(self, test_method):
        FakeModuleBuilder.reset()
        cleanup_moksha()
        for i in range(20):
            try:
                os.remove(build_logs.path(i))
            except Exception:
                pass

    @pytest.mark.parametrize('mmd_version', [1, 2])
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build(self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc,
                          mmd_version):
        """
        Tests the build of testmodule.yaml using FakeModuleBuilder which
        succeeds everytime.
        """
        if mmd_version == 1:
            yaml_file = 'testmodule.yaml'
        else:
            yaml_file = 'testmodule_v2.yaml'
        FakeSCM(mocked_scm, 'testmodule', yaml_file,
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        data = json.loads(rv.data)
        module_build_id = data['id']

        # Check that components are tagged after the batch is built.
        tag_groups = []
        tag_groups.append(set(['perl-Tangerine-1-1', 'perl-List-Compare-1-1']))
        tag_groups.append(set(['tangerine-1-1']))

        def on_tag_artifacts_cb(cls, artifacts, dest_tag=True):
            assert tag_groups.pop(0) == set(artifacts)

        FakeModuleBuilder.on_tag_artifacts_cb = on_tag_artifacts_cb

        # Check that the components are added to buildroot after the batch
        # is built.
        buildroot_groups = []
        buildroot_groups.append(set(['module-build-macros-1-1']))
        buildroot_groups.append(set(['perl-Tangerine-1-1', 'perl-List-Compare-1-1']))
        buildroot_groups.append(set(['tangerine-1-1']))

        def on_buildroot_add_artifacts_cb(cls, artifacts, install):
            assert buildroot_groups.pop(0) == set(artifacts)

        FakeModuleBuilder.on_buildroot_add_artifacts_cb = on_buildroot_add_artifacts_cb

        msgs = []
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)

        # All components should be built and module itself should be in "done"
        # or "ready" state.
        for build in models.ComponentBuild.query.filter_by(module_id=module_build_id).all():
            assert build.state == koji.BUILD_STATES['COMPLETE']
            assert build.module_build.state in [models.BUILD_STATES["done"],
                                                models.BUILD_STATES["ready"]]

        # All components has to be tagged, so tag_groups and buildroot_groups are empty...
        assert tag_groups == []
        assert buildroot_groups == []
        module_build = models.ModuleBuild.query.get(module_build_id)
        assert module_build.module_builds_trace[0].state == models.BUILD_STATES['init']
        assert module_build.module_builds_trace[1].state == models.BUILD_STATES['wait']
        assert module_build.module_builds_trace[2].state == models.BUILD_STATES['build']
        assert module_build.module_builds_trace[3].state == models.BUILD_STATES['done']
        assert module_build.module_builds_trace[4].state == models.BUILD_STATES['ready']
        assert len(module_build.module_builds_trace) == 5

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_no_components(self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc):
        """
        Tests the build of a module with no components
        """
        FakeSCM(mocked_scm, 'python3', 'python3-no-components.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        data = json.loads(rv.data)
        module_build_id = data['id']
        msgs = []
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)

        module_build = models.ModuleBuild.query.filter_by(id=module_build_id).one()
        # Make sure no component builds were registered
        assert len(module_build.component_builds) == 0
        # Make sure the build is done
        assert module_build.state == models.BUILD_STATES['ready']

    @patch('module_build_service.config.Config.check_for_eol',
           new_callable=PropertyMock, return_value=True)
    @patch('module_build_service.utils.submit._is_eol_in_pdc', return_value=True)
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_eol_module(self, mocked_scm, mocked_get_user, is_eol, check,
                                     conf_system, dbg, hmsc):
        """ Tests the build of a module with an eol stream.  """
        FakeSCM(mocked_scm, 'python3', 'python3-no-components.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        assert rv.status_code == 400
        data = json.loads(rv.data)
        assert data['status'] == 400
        assert data['message'] == u'Module python3:master is marked as EOL in PDC.'

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_from_yaml_not_allowed(
            self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc):
        FakeSCM(mocked_scm, "testmodule", "testmodule.yaml")

        testmodule = os.path.join(base_dir, 'staged_data', 'testmodule.yaml')
        with open(testmodule) as f:
            yaml = f.read()

        with patch.object(module_build_service.config.Config, 'yaml_submit_allowed',
                          new_callable=PropertyMock, return_value=False):
            rv = self.client.post('/module-build-service/1/module-builds/',
                                  content_type='multipart/form-data',
                                  data={'yaml': (testmodule, yaml)})
            data = json.loads(rv.data)
            assert data['status'] == 403
            assert data['message'] == 'YAML submission is not enabled'

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_from_yaml_allowed(
            self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc):
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        testmodule = os.path.join(base_dir, 'staged_data', 'testmodule.yaml')

        with patch.object(module_build_service.config.Config, 'yaml_submit_allowed',
                          new_callable=PropertyMock, return_value=True):
            with open(testmodule, 'rb') as f:
                yaml_file = FileStorage(f)
                rv = self.client.post('/module-build-service/1/module-builds/',
                                      content_type='multipart/form-data',
                                      data={'yaml': yaml_file})
            data = json.loads(rv.data)
            assert data['id'] == 2
        msgs = []
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)
        assert models.ModuleBuild.query.first().state == models.BUILD_STATES['ready']

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_cancel(self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc):
        """
        Submit all builds for a module and cancel the module build later.
        """
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        data = json.loads(rv.data)
        module_build_id = data['id']

        # This callback is called before return of FakeModuleBuilder.build()
        # method. We just cancel the build here using the web API to simulate
        # user cancelling the build in the middle of building.
        def on_build_cb(cls, artifact_name, source):
            self.client.patch('/module-build-service/1/module-builds/' + str(module_build_id),
                              data=json.dumps({'state': 'failed'}))

        cancelled_tasks = []

        def on_cancel_cb(cls, task_id):
            cancelled_tasks.append(task_id)

        # We do not want the builds to COMPLETE, but instead we want them
        # to be in the BULDING state after the FakeModuleBuilder.build().
        FakeModuleBuilder.BUILD_STATE = "BUILDING"
        FakeModuleBuilder.on_build_cb = on_build_cb
        FakeModuleBuilder.on_cancel_cb = on_cancel_cb

        msgs = []
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)

        # Because we did not finished single component build and canceled the
        # module build, all components and even the module itself should be in
        # failed state with state_reason se to cancellation message.
        for build in models.ComponentBuild.query.filter_by(module_id=module_build_id).all():
            assert build.state == koji.BUILD_STATES['FAILED']
            assert build.state_reason == "Canceled by Homer J. Simpson."
            assert build.module_build.state == models.BUILD_STATES["failed"]
            assert build.module_build.state_reason == "Canceled by Homer J. Simpson."

            # Check that cancel_build has been called for this build
            if build.task_id:
                assert build.task_id in cancelled_tasks

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_instant_complete(
            self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc):
        """
        Tests the build of testmodule.yaml using FakeModuleBuilder which
        succeeds everytime.
        """
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        data = json.loads(rv.data)
        module_build_id = data['id']
        FakeModuleBuilder.INSTANT_COMPLETE = True

        msgs = []
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)

        # All components should be built and module itself should be in "done"
        # or "ready" state.
        for build in models.ComponentBuild.query.filter_by(module_id=module_build_id).all():
            assert build.state == koji.BUILD_STATES['COMPLETE']
            assert build.module_build.state in [models.BUILD_STATES["done"],
                                                models.BUILD_STATES["ready"]]

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.num_concurrent_builds",
           new_callable=PropertyMock, return_value=1)
    def test_submit_build_concurrent_threshold(self, conf_num_concurrent_builds,
                                               mocked_scm, mocked_get_user,
                                               conf_system, dbg, hmsc):
        """
        Tests the build of testmodule.yaml using FakeModuleBuilder with
        num_concurrent_builds set to 1.
        """
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        data = json.loads(rv.data)
        module_build_id = data['id']

        def stop(message):
            """
            Stop the scheduler when the module is built or when we try to build
            more components than the num_concurrent_builds.
            """
            main_stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
            over_threshold = conf.num_concurrent_builds < \
                db.session.query(models.ComponentBuild).filter_by(
                    state=koji.BUILD_STATES['BUILDING']).count()
            return main_stop(message) or over_threshold

        msgs = []
        module_build_service.scheduler.main(msgs, stop)

        # All components should be built and module itself should be in "done"
        # or "ready" state.
        for build in models.ComponentBuild.query.filter_by(module_id=module_build_id).all():
            assert build.state == koji.BUILD_STATES['COMPLETE']
            # When this fails, it can mean that num_concurrent_builds
            # threshold has been met.
            assert build.module_build.state in [models.BUILD_STATES["done"],
                                                models.BUILD_STATES["ready"]]

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.num_concurrent_builds",
           new_callable=PropertyMock, return_value=2)
    def test_try_to_reach_concurrent_threshold(self, conf_num_concurrent_builds,
                                               mocked_scm, mocked_get_user,
                                               conf_system, dbg, hmsc):
        """
        Tests that we try to submit new component build right after
        the previous one finished without waiting for all
        the num_concurrent_builds to finish.
        """
        FakeSCM(mocked_scm, 'testmodule-more-components', 'testmodule-more-components.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        # Holds the number of concurrent component builds during
        # the module build.
        TestBuild._global_var = []

        def stop(message):
            """
            Stop the scheduler when the module is built or when we try to build
            more components than the num_concurrent_builds.
            """
            main_stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
            num_building = db.session.query(models.ComponentBuild).filter_by(
                state=koji.BUILD_STATES['BUILDING']).count()
            over_threshold = conf.num_concurrent_builds < num_building
            TestBuild._global_var.append(num_building)
            return main_stop(message) or over_threshold

        msgs = []
        module_build_service.scheduler.main(msgs, stop)

        # _global_var looks similar to this: [0, 1, 0, 0, 2, 2, 1, 0, 0, 0]
        # It shows the number of concurrent builds in the time. At first we
        # want to remove adjacent duplicate entries, because we only care
        # about changes.
        # We are building two batches, so there should be just two situations
        # when we should be building just single component:
        #   1) module-base-macros in first batch.
        #   2) The last component of second batch.
        # If we are building single component more often, num_concurrent_builds
        # does not work correctly.
        num_builds = [k for k, g in itertools.groupby(TestBuild._global_var)]
        assert num_builds.count(1) == 2

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.num_concurrent_builds",
           new_callable=PropertyMock, return_value=1)
    def test_build_in_batch_fails(self, conf_num_concurrent_builds, mocked_scm,
                                  mocked_get_user, conf_system, dbg, hmsc):
        """
        Tests that if the build in batch fails, other components in a batch
        are still build, but next batch is not started.
        """
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        data = json.loads(rv.data)
        module_build_id = data['id']

        def on_build_cb(cls, artifact_name, source):
            # fail perl-Tangerine build
            if artifact_name.startswith("perl-Tangerine"):
                FakeModuleBuilder.BUILD_STATE = "FAILED"
            else:
                FakeModuleBuilder.BUILD_STATE = "COMPLETE"

        FakeModuleBuilder.on_build_cb = on_build_cb

        # Check that no components are tagged when single component fails
        # in batch.
        def on_tag_artifacts_cb(cls, artifacts, dest_tag=True):
            raise ValueError("No component should be tagged.")
        FakeModuleBuilder.on_tag_artifacts_cb = on_tag_artifacts_cb

        msgs = []
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)

        for c in models.ComponentBuild.query.filter_by(module_id=module_build_id).all():
            # perl-Tangerine is expected to fail as configured in on_build_cb.
            if c.package == "perl-Tangerine":
                assert c.state == koji.BUILD_STATES['FAILED']
            # tangerine is expected to fail, because it is in batch 3, but
            # we had a failing component in batch 2.
            elif c.package == "tangerine":
                assert c.state == koji.BUILD_STATES['FAILED']
                assert c.state_reason == "Some components failed to build."
            else:
                assert c.state == koji.BUILD_STATES['COMPLETE']

            # Whole module should be failed.
            assert c.module_build.state == models.BUILD_STATES['failed']
            assert c.module_build.state_reason == "Some components failed to build."

            # We should end up with batch 2 and never start batch 3, because
            # there were failed components in batch 2.
            assert c.module_build.batch == 2

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.num_concurrent_builds",
           new_callable=PropertyMock, return_value=1)
    def test_all_builds_in_batch_fail(self, conf_num_concurrent_builds, mocked_scm,
                                      mocked_get_user, conf_system, dbg, hmsc):
        """
        Tests that if the build in batch fails, other components in a batch
        are still build, but next batch is not started.
        """
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        data = json.loads(rv.data)
        module_build_id = data['id']

        def on_build_cb(cls, artifact_name, source):
            # Next components *after* the module-build-macros will fail
            # to build.
            if not artifact_name.startswith("module-build-macros"):
                FakeModuleBuilder.BUILD_STATE = "FAILED"

        FakeModuleBuilder.on_build_cb = on_build_cb

        msgs = []
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)

        for c in models.ComponentBuild.query.filter_by(module_id=module_build_id).all():
            # perl-Tangerine is expected to fail as configured in on_build_cb.
            if c.package == "module-build-macros":
                assert c.state == koji.BUILD_STATES['COMPLETE']
            else:
                assert c.state == koji.BUILD_STATES['FAILED']

            # Whole module should be failed.
            assert c.module_build.state == models.BUILD_STATES['failed']
            assert c.module_build.state_reason == "Some components failed to build."

            # We should end up with batch 2 and never start batch 3, because
            # there were failed components in batch 2.
            assert c.module_build.batch == 2

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_reuse_all(self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc):
        """
        Tests that we do not try building module-build-macros when reusing all
        components in a module build.
        """
        reuse_component_init_data()

        def on_build_cb(cls, artifact_name, source):
            raise ValueError("All components should be reused, not build.")
        FakeModuleBuilder.on_build_cb = on_build_cb

        # Check that components are tagged after the batch is built.
        tag_groups = []
        tag_groups.append(set(
            ['perl-Tangerine-0.23-1.module+0+d027b723',
             'perl-List-Compare-0.53-5.module+0+d027b723',
             'tangerine-0.22-3.module+0+d027b723']))

        def on_tag_artifacts_cb(cls, artifacts, dest_tag=True):
            if dest_tag is True:
                assert tag_groups.pop(0) == set(artifacts)
        FakeModuleBuilder.on_tag_artifacts_cb = on_tag_artifacts_cb

        buildtag_groups = []
        buildtag_groups.append(set(
            ['perl-Tangerine-0.23-1.module+0+d027b723',
             'perl-List-Compare-0.53-5.module+0+d027b723',
             'tangerine-0.22-3.module+0+d027b723']))

        def on_buildroot_add_artifacts_cb(cls, artifacts, install):
            assert buildtag_groups.pop(0) == set(artifacts)
        FakeModuleBuilder.on_buildroot_add_artifacts_cb = on_buildroot_add_artifacts_cb

        msgs = [MBSModule("local module build", 3, 1)]
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)

        reused_component_ids = {"module-build-macros": None, "tangerine": 3,
                                "perl-Tangerine": 1, "perl-List-Compare": 2}

        # All components should be built and module itself should be in "done"
        # or "ready" state.
        for build in models.ComponentBuild.query.filter_by(module_id=3).all():
            assert build.state == koji.BUILD_STATES['COMPLETE']
            assert build.module_build.state in [models.BUILD_STATES["done"],
                                                models.BUILD_STATES["ready"]]
            assert build.reused_component_id == reused_component_ids[build.package]

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_reuse_all_without_build_macros(self, mocked_scm, mocked_get_user,
                                                         conf_system, dbg, hmsc):
        """
        Tests that we can reuse components even when the reused module does
        not have module-build-macros component.
        """
        reuse_component_init_data()

        models.ComponentBuild.query.filter_by(package="module-build-macros").delete()
        assert len(models.ComponentBuild.query.filter_by(package="module-build-macros").all()) == 0

        db.session.commit()

        def on_build_cb(cls, artifact_name, source):
            raise ValueError("All components should be reused, not build.")
        FakeModuleBuilder.on_build_cb = on_build_cb

        # Check that components are tagged after the batch is built.
        tag_groups = []
        tag_groups.append(set(
            ['perl-Tangerine-0.23-1.module+0+d027b723',
             'perl-List-Compare-0.53-5.module+0+d027b723',
             'tangerine-0.22-3.module+0+d027b723']))

        def on_tag_artifacts_cb(cls, artifacts, dest_tag=True):
            if dest_tag is True:
                assert tag_groups.pop(0) == set(artifacts)
        FakeModuleBuilder.on_tag_artifacts_cb = on_tag_artifacts_cb

        buildtag_groups = []
        buildtag_groups.append(set(
            ['perl-Tangerine-0.23-1.module+0+d027b723',
             'perl-List-Compare-0.53-5.module+0+d027b723',
             'tangerine-0.22-3.module+0+d027b723']))

        def on_buildroot_add_artifacts_cb(cls, artifacts, install):
            assert buildtag_groups.pop(0) == set(artifacts)
        FakeModuleBuilder.on_buildroot_add_artifacts_cb = on_buildroot_add_artifacts_cb

        msgs = [MBSModule("local module build", 3, 1)]
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)

        # All components should be built and module itself should be in "done"
        # or "ready" state.
        for build in models.ComponentBuild.query.filter_by(module_id=3).all():
            assert build.state == koji.BUILD_STATES['COMPLETE']
            assert build.module_build.state in [models.BUILD_STATES["done"],
                                                models.BUILD_STATES["ready"]]
            assert build.package != "module-build-macros"

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_resume(self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc):
        """
        Tests that resuming the build works even when previous batches
        are already built.
        """
        now = datetime.utcnow()
        submitted_time = now - timedelta(minutes=3)
        # Create a module in the failed state
        build_one = models.ModuleBuild()
        build_one.name = 'testmodule'
        build_one.stream = 'master'
        build_one.version = '2820180205135154'
        build_one.build_context = 'return_runtime_context'
        build_one.ref_build_context = 'return_runtime_context'
        build_one.runtime_context = '9c690d0e'
        build_one.context = '9c690d0e'
        build_one.state = models.BUILD_STATES['failed']
        current_dir = os.path.dirname(__file__)
        formatted_testmodule_yml_path = os.path.join(
            current_dir, '..', 'staged_data', 'formatted_testmodule.yaml')
        with open(formatted_testmodule_yml_path, 'r') as f:
            build_one.modulemd = f.read()
        build_one.koji_tag = 'module-testmodule-master-20180205135154-9c690d0e'
        build_one.scmurl = 'git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#7fea453'
        build_one.batch = 2
        build_one.owner = 'Homer J. Simpson'
        build_one.time_submitted = submitted_time
        build_one.time_modified = now
        build_one.rebuild_strategy = 'changed-and-after'
        # It went from init, to wait, to build, and then failed
        mbt_one = models.ModuleBuildTrace(
            state_time=submitted_time, state=models.BUILD_STATES['init'])
        mbt_two = models.ModuleBuildTrace(
            state_time=now - timedelta(minutes=2), state=models.BUILD_STATES['wait'])
        mbt_three = models.ModuleBuildTrace(
            state_time=now - timedelta(minutes=1), state=models.BUILD_STATES['build'])
        mbt_four = models.ModuleBuildTrace(state_time=now, state=build_one.state)
        build_one.module_builds_trace.append(mbt_one)
        build_one.module_builds_trace.append(mbt_two)
        build_one.module_builds_trace.append(mbt_three)
        build_one.module_builds_trace.append(mbt_four)
        # Successful component
        component_one = models.ComponentBuild()
        component_one.package = 'perl-Tangerine'
        component_one.format = 'rpms'
        component_one.scmurl = 'git://pkgs.stg.fedoraproject.org/rpms/perl-Tangerine.git?#master'
        component_one.state = koji.BUILD_STATES['COMPLETE']
        component_one.nvr = 'perl-Tangerine-0:0.22-2.module+0+d027b723'
        component_one.batch = 2
        component_one.module_id = 2
        component_one.ref = '7e96446223f1ad84a26c7cf23d6591cd9f6326c6'
        component_one.tagged = True
        component_one.tagged_in_final = True
        # Failed component
        component_two = models.ComponentBuild()
        component_two.package = 'perl-List-Compare'
        component_two.format = 'rpms'
        component_two.scmurl = \
            'git://pkgs.stg.fedoraproject.org/rpms/perl-List-Compare.git?#master'
        component_two.state = koji.BUILD_STATES['FAILED']
        component_two.batch = 2
        component_two.module_id = 2
        # Component that isn't started yet
        component_three = models.ComponentBuild()
        component_three.package = 'tangerine'
        component_three.format = 'rpms'
        component_three.scmurl = 'git://pkgs.stg.fedoraproject.org/rpms/tangerine.git?#master'
        component_three.batch = 3
        component_three.module_id = 2
        # module-build-macros
        component_four = models.ComponentBuild()
        component_four.package = 'module-build-macros'
        component_four.format = 'rpms'
        component_four.state = koji.BUILD_STATES['COMPLETE']
        component_four.scmurl = (
            '/tmp/module_build_service-build-macrosqr4AWH/SRPMS/module-build-macros-0.1-1.'
            'module_testmodule_master_20170109091357.src.rpm')
        component_four.batch = 1
        component_four.module_id = 2
        component_four.tagged = True
        component_four.build_time_only = True

        db.session.add(build_one)
        db.session.add(component_one)
        db.session.add(component_two)
        db.session.add(component_three)
        db.session.add(component_four)
        db.session.commit()
        db.session.expire_all()

        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        # Resubmit the failed module
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        data = json.loads(rv.data)
        module_build_id = data['id']
        module_build = models.ModuleBuild.query.filter_by(id=module_build_id).one()
        components = models.ComponentBuild.query.filter_by(
            module_id=module_build_id, batch=2).order_by(models.ComponentBuild.id).all()
        # Make sure the build went from failed to wait
        assert module_build.state == models.BUILD_STATES['wait']
        assert module_build.state_reason == 'Resubmitted by Homer J. Simpson'
        # Make sure the state was reset on the failed component
        assert components[1].state is None
        db.session.expire_all()

        # Run the backend
        msgs = []
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)

        # All components should be built and module itself should be in "done"
        # or "ready" state.
        for build in models.ComponentBuild.query.filter_by(module_id=module_build_id).all():
            assert build.state == koji.BUILD_STATES['COMPLETE']
            assert build.module_build.state in [models.BUILD_STATES['done'],
                                                models.BUILD_STATES['ready']]

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_resume_recover_orphaned_macros(
            self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc):
        """
        Tests that resuming the build works when module-build-macros is orphaned but marked as
        failed in the database
        """
        FakeModuleBuilder.INSTANT_COMPLETE = True
        now = datetime.utcnow()
        submitted_time = now - timedelta(minutes=3)
        # Create a module in the failed state
        build_one = models.ModuleBuild()
        build_one.name = 'testmodule'
        build_one.stream = 'master'
        build_one.version = '2820180205135154'
        build_one.build_context = 'return_runtime_context'
        build_one.ref_build_context = 'return_runtime_context'
        build_one.runtime_context = '9c690d0e'
        build_one.state = models.BUILD_STATES['failed']
        # this is not calculated by real but just a value to
        # match the calculated context from expanded test mmd
        build_one.context = '9c690d0e'
        current_dir = os.path.dirname(__file__)
        formatted_testmodule_yml_path = os.path.join(
            current_dir, '..', 'staged_data', 'formatted_testmodule.yaml')
        with open(formatted_testmodule_yml_path, 'r') as f:
            build_one.modulemd = f.read()
        build_one.koji_tag = 'module-testmodule-master-20180205135154-6ef9a711'
        build_one.scmurl = 'git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#7fea453'
        build_one.batch = 2
        build_one.owner = 'Homer J. Simpson'
        build_one.time_submitted = submitted_time
        build_one.time_modified = now
        build_one.rebuild_strategy = 'changed-and-after'
        # It went from init, to wait, to build, and then failed
        mbt_one = models.ModuleBuildTrace(
            state_time=submitted_time, state=models.BUILD_STATES['init'])
        mbt_two = models.ModuleBuildTrace(
            state_time=now - timedelta(minutes=2), state=models.BUILD_STATES['wait'])
        mbt_three = models.ModuleBuildTrace(
            state_time=now - timedelta(minutes=1), state=models.BUILD_STATES['build'])
        mbt_four = models.ModuleBuildTrace(state_time=now, state=build_one.state)
        build_one.module_builds_trace.append(mbt_one)
        build_one.module_builds_trace.append(mbt_two)
        build_one.module_builds_trace.append(mbt_three)
        build_one.module_builds_trace.append(mbt_four)
        # Components that haven't started yet
        component_one = models.ComponentBuild()
        component_one.package = 'perl-Tangerine'
        component_one.format = 'rpms'
        component_one.scmurl = 'git://pkgs.stg.fedoraproject.org/rpms/perl-Tangerine.git?#master'
        component_one.batch = 2
        component_one.module_id = 2
        component_two = models.ComponentBuild()
        component_two.package = 'perl-List-Compare'
        component_two.format = 'rpms'
        component_two.scmurl = \
            'git://pkgs.stg.fedoraproject.org/rpms/perl-List-Compare.git?#master'
        component_two.batch = 2
        component_two.module_id = 2
        component_three = models.ComponentBuild()
        component_three.package = 'tangerine'
        component_three.format = 'rpms'
        component_three.scmurl = 'git://pkgs.stg.fedoraproject.org/rpms/tangerine.git?#master'
        component_three.batch = 3
        component_three.module_id = 2
        # Failed module-build-macros
        component_four = models.ComponentBuild()
        component_four.package = 'module-build-macros'
        component_four.format = 'rpms'
        component_four.state = koji.BUILD_STATES['FAILED']
        component_four.scmurl = (
            '/tmp/module_build_service-build-macrosqr4AWH/SRPMS/module-build-macros-0.1-1.'
            'module_testmodule_master_20180205135154.src.rpm')
        component_four.batch = 1
        component_four.module_id = 2
        component_four.build_time_only = True

        db.session.add(build_one)
        db.session.add(component_one)
        db.session.add(component_two)
        db.session.add(component_three)
        db.session.add(component_four)
        db.session.commit()
        db.session.expire_all()

        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml', '7fea453')
        # Resubmit the failed module
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#7fea453'}))

        data = json.loads(rv.data)
        module_build_id = data['id']
        module_build = models.ModuleBuild.query.filter_by(id=module_build_id).one()
        # Make sure the build went from failed to wait
        assert module_build.state == models.BUILD_STATES['wait']
        assert module_build.state_reason == 'Resubmitted by Homer J. Simpson'
        # Make sure the state was reset on the failed component
        for c in module_build.component_builds:
            assert c.state is None
        db.session.expire_all()

        # Run the backend
        msgs = []
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main(msgs, stop)

        # All components should be built and module itself should be in "done"
        # or "ready" state.
        for build in models.ComponentBuild.query.filter_by(module_id=module_build_id).all():
            assert build.state == koji.BUILD_STATES['COMPLETE']
            assert build.module_build.state in [models.BUILD_STATES['done'],
                                                models.BUILD_STATES['ready']]

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_resume_failed_init(
            self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc):
        """
        Tests that resuming the build works when the build failed during the init step
        """
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)

        with patch('module_build_service.utils.submit.format_mmd') as mock_format_mmd:
            mock_format_mmd.side_effect = Forbidden(
                'Custom component repositories aren\'t allowed.')
            rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
                {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                    'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))
            # Run the backend so that it fails in the "init" handler
            module_build_service.scheduler.main([], stop)
            cleanup_moksha()

        module_build_id = json.loads(rv.data)['id']
        module_build = models.ModuleBuild.query.filter_by(id=module_build_id).one()
        assert module_build.state == models.BUILD_STATES['failed']
        assert module_build.state_reason == 'Custom component repositories aren\'t allowed.'
        assert len(module_build.module_builds_trace) == 2
        assert module_build.module_builds_trace[0].state == models.BUILD_STATES['init']
        assert module_build.module_builds_trace[1].state == models.BUILD_STATES['failed']

        # Resubmit the failed module
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master',
             'scmurl': ('git://pkgs.stg.fedoraproject.org/modules/testmodule.git?'
                        '#620ec77321b2ea7b0d67d82992dda3e1d67055b4')}))

        module_build = models.ModuleBuild.query.filter_by(id=module_build_id).one()
        components = models.ComponentBuild.query.filter_by(
            module_id=module_build_id, batch=2).order_by(models.ComponentBuild.id).all()
        # Make sure the build went from failed to init
        assert module_build.state == models.BUILD_STATES['init']
        assert module_build.state_reason == 'Resubmitted by Homer J. Simpson'
        # Make sure there are no components
        assert components == []
        db.session.expire_all()

        # Run the backend again
        module_build_service.scheduler.main([], stop)

        # All components should be built and module itself should be in "done"
        # or "ready" state.
        for build in models.ComponentBuild.query.filter_by(module_id=module_build_id).all():
            assert build.state == koji.BUILD_STATES['COMPLETE']
            assert build.module_build.state in [models.BUILD_STATES['done'],
                                                models.BUILD_STATES['ready']]

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_resume_init_fail(
            self, mocked_scm, mocked_get_user, conf_system, dbg, hmsc):
        """
        Tests that resuming the build fails when the build is in init state
        """
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')
        # Post so a module is in the init phase
        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))
        assert rv.status_code == 201
        # Run the backend
        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)
        module_build_service.scheduler.main([], stop)
        # Post again and make sure it fails
        rv2 = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))
        data = json.loads(rv2.data)
        expected = {
            'error': 'Conflict',
            'message': ('Module (state=5) already exists. Only a new build or resubmission of a '
                        'failed build is allowed.'),
            'status': 409
        }
        assert data == expected

    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    def test_submit_build_repo_regen_not_started_batch(self, mocked_scm, mocked_get_user,
                                                       conf_system, dbg, hmsc):
        """
        Tests that if MBS starts a new batch, the concurrent component threshold is met before a
        build can start, and an unexpected repo regen occurs, the build will not fail.

        See: https://pagure.io/fm-orchestrator/issue/864
        """
        FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

        rv = self.client.post('/module-build-service/1/module-builds/', data=json.dumps(
            {'branch': 'master', 'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

        data = json.loads(rv.data)
        module_build_id = data['id']

        def _stop_condition(message):
            # Stop the backend if the module batch is 2 (where we simulate the concurrent threshold
            # being met). For safety, also stop the backend if the module erroneously completes.
            module = db.session.query(models.ModuleBuild).get(module_build_id)
            return module.batch == 2 or module.state >= models.BUILD_STATES['done']

        with patch('module_build_service.utils.batches.at_concurrent_component_threshold') as \
                mock_acct:
            # Once we get to batch 2, then simulate the concurrent threshold being met
            def _at_concurrent_component_threshold(config, session):
                return db.session.query(models.ModuleBuild).get(module_build_id).batch == 2
            mock_acct.side_effect = _at_concurrent_component_threshold
            module_build_service.scheduler.main([], _stop_condition)

        # Only module-build-macros should be built
        for build in db.session.query(models.ComponentBuild).filter_by(
                module_id=module_build_id).all():
            if build.package == 'module-build-macros':
                assert build.state == koji.BUILD_STATES['COMPLETE']
            else:
                assert build.state is None
            assert build.module_build.state == models.BUILD_STATES['build']

        # Simulate a random repo regen message that MBS didn't expect
        cleanup_moksha()
        module = db.session.query(models.ModuleBuild).get(module_build_id)
        msgs = [module_build_service.messaging.KojiRepoChange(
            msg_id='a faked internal message', repo_tag=module.koji_tag + '-build')]
        db.session.expire_all()
        # Stop after processing the seeded message
        module_build_service.scheduler.main(msgs, lambda message: True)
        # Make sure the module build didn't fail so that the poller can resume it later
        module = db.session.query(models.ModuleBuild).get(module_build_id)
        assert module.state == models.BUILD_STATES['build']


@patch("module_build_service.config.Config.system",
       new_callable=PropertyMock, return_value="testlocal")
class TestLocalBuild:

    def setup_method(self, test_method):
        FakeModuleBuilder.on_build_cb = None
        FakeModuleBuilder.backend = 'testlocal'
        GenericBuilder.register_backend_class(FakeModuleBuilder)
        self.client = app.test_client()
        clean_database()

    def teardown_method(self, test_method):
        FakeModuleBuilder.reset()
        cleanup_moksha()
        for i in range(20):
            try:
                os.remove(build_logs.path(i))
            except Exception:
                pass

    @patch('module_build_service.utils.submit.record_stream_collision_modules')
    @patch('module_build_service.auth.get_user', return_value=user)
    @patch('module_build_service.scm.SCM')
    @patch("module_build_service.config.Config.mock_resultsdir",
           new_callable=PropertyMock,
           return_value=path.join(base_dir, 'staged_data', "local_builds"))
    def test_submit_build_local_dependency(
            self, resultsdir, mocked_scm, mocked_get_user, conf_system, hmsc):
        """
        Tests local module build dependency.
        """
        with app.app_context():
            module_build_service.utils.load_local_builds(["platform"])
            FakeSCM(mocked_scm, 'testmodule', 'testmodule.yaml',
                    '620ec77321b2ea7b0d67d82992dda3e1d67055b4')

            rv = self.client.post(
                '/module-build-service/1/module-builds/', data=json.dumps(
                    {'branch': 'master',
                     'scmurl': 'git://pkgs.stg.fedoraproject.org/modules/'
                     'testmodule.git?#620ec77321b2ea7b0d67d82992dda3e1d67055b4'}))

            data = json.loads(rv.data)
            module_build_id = data['id']

            # Local base-runtime has changed profiles, so we can detect we use
            # the local one and not the main one.
            FakeModuleBuilder.DEFAULT_GROUPS = {
                'srpm-build':
                    set(['bar']),
                'build':
                    set(['foo'])}

            msgs = []
            stop = module_build_service.scheduler.make_simple_stop_condition(
                db.session)
            module_build_service.scheduler.main(msgs, stop)

            # All components should be built and module itself should be in "done"
            # or "ready" state.
            for build in models.ComponentBuild.query.filter_by(
                    module_id=module_build_id).all():
                assert build.state == koji.BUILD_STATES['COMPLETE']
                assert build.module_build.state in [models.BUILD_STATES["done"],
                                                    models.BUILD_STATES["ready"]]



# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Matt Prahl <mprahl@redhat.com>

import os

from datetime import datetime
from mock import patch, PropertyMock
import pytest

import module_build_service.resolver as mbs_resolver
from module_build_service import app, db, models, glib, utils, Modulemd
from module_build_service.utils import import_mmd, load_mmd
from module_build_service.models import ModuleBuild
import tests


base_dir = os.path.join(os.path.dirname(__file__), "..")


class TestDBModule:

    def setup_method(self):
        tests.reuse_component_init_data()

    def test_get_buildrequired_modulemds(self):
        mmd = load_mmd(os.path.join(base_dir, 'staged_data', 'platform.yaml'), True)
        mmd.set_stream('f30.1.3')
        import_mmd(db.session, mmd)
        platform_f300103 = ModuleBuild.query.filter_by(stream='f30.1.3').one()
        mmd.set_name("testmodule")
        mmd.set_stream("master")
        mmd.set_version(20170109091357)
        mmd.set_context("123")
        build = ModuleBuild(
            name='testmodule',
            stream='master',
            version=20170109091357,
            state=5,
            build_context='dd4de1c346dcf09ce77d38cd4e75094ec1c08ec3',
            runtime_context='ec4de1c346dcf09ce77d38cd4e75094ec1c08ef7',
            context='7c29193d',
            koji_tag='module-testmodule-master-20170109091357-7c29193d',
            scmurl='git://pkgs.stg.fedoraproject.org/modules/testmodule.git?#ff1ea79',
            batch=3,
            owner='Dr. Pepper',
            time_submitted=datetime(2018, 11, 15, 16, 8, 18),
            time_modified=datetime(2018, 11, 15, 16, 19, 35),
            rebuild_strategy='changed-and-after',
            modulemd=mmd.dumps()
        )
        build.buildrequires.append(platform_f300103)
        db.session.add(build)
        db.session.commit()

        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='db')
        result = resolver.get_buildrequired_modulemds(
            "testmodule", "master", platform_f300103.mmd().dup_nsvc())
        nsvcs = set([m.dup_nsvc() for m in result])
        assert nsvcs == set(['testmodule:master:20170109091357:123'])

    @pytest.mark.parametrize('stream_versions', [False, True])
    def test_get_module_modulemds_stream_versions(self, stream_versions):
        tests.init_data(1, multiple_stream_versions=True)
        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='db')
        result = resolver.get_module_modulemds(
            "platform", "f29.1.0", stream_version_lte=stream_versions)
        nsvcs = set([mmd.dup_nsvc() for mmd in result])
        if stream_versions:
            assert nsvcs == set(['platform:f29.1.0:3:00000000', 'platform:f29.0.0:3:00000000'])
        else:
            assert nsvcs == set(['platform:f29.1.0:3:00000000'])

    @pytest.mark.parametrize('empty_buildrequires', [False, True])
    def test_get_module_build_dependencies(self, empty_buildrequires):
        """
        Tests that the buildrequires of testmodule are returned
        """
        expected = set(['module-f28-build'])
        module = models.ModuleBuild.query.get(2)
        if empty_buildrequires:
            expected = set()
            module = models.ModuleBuild.query.get(2)
            mmd = module.mmd()
            # Wipe out the dependencies
            mmd.set_dependencies()
            xmd = glib.from_variant_dict(mmd.get_xmd())
            xmd['mbs']['buildrequires'] = {}
            mmd.set_xmd(glib.dict_values(xmd))
            module.modulemd = mmd.dumps()
            db.session.add(module)
            db.session.commit()
        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='db')
        result = resolver.get_module_build_dependencies(
            'testmodule', 'master', '20170109091357', '78e4a6fd').keys()
        assert set(result) == expected

    def test_get_module_build_dependencies_recursive(self):
        """
        Tests that the buildrequires are returned when it is two layers deep
        """
        # Add testmodule2 that requires testmodule
        module = models.ModuleBuild.query.get(3)
        mmd = module.mmd()
        mmd.set_name('testmodule2')
        mmd.set_version(20180123171545)
        requires = mmd.get_dependencies()[0].get_requires()
        requires['testmodule'] = Modulemd.SimpleSet()
        requires['testmodule'].add('master')
        mmd.get_dependencies()[0].set_requires(requires)
        xmd = glib.from_variant_dict(mmd.get_xmd())
        xmd['mbs']['requires']['testmodule'] = {
            'filtered_rpms': [],
            'ref': '620ec77321b2ea7b0d67d82992dda3e1d67055b4',
            'stream': 'master',
            'version': '20180205135154'
        }
        mmd.set_xmd(glib.dict_values(xmd))
        module.modulemd = mmd.dumps()
        module.name = 'testmodule2'
        module.version = str(mmd.get_version())
        module.koji_tag = 'module-ae2adf69caf0e1b6'

        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='db')
        result = resolver.get_module_build_dependencies(
            'testmodule2', 'master', '20180123171545', 'c40c156c').keys()
        assert set(result) == set(['module-f28-build'])

    @patch("module_build_service.config.Config.system",
           new_callable=PropertyMock, return_value="test")
    @patch("module_build_service.config.Config.mock_resultsdir",
           new_callable=PropertyMock,
           return_value=os.path.join(base_dir, 'staged_data', "local_builds"))
    def test_get_module_build_dependencies_recursive_requires(
            self, resultdir, conf_system):
        """
        Tests that it returns the requires of the buildrequires recursively
        """
        with app.app_context():
            utils.load_local_builds(["platform", "parent", "child", "testmodule"])

            build = models.ModuleBuild.local_modules(
                db.session, "child", "master")
            resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='db')
            result = resolver.get_module_build_dependencies(mmd=build[0].mmd()).keys()

            local_path = os.path.join(base_dir, 'staged_data', "local_builds")

            expected = [
                os.path.join(
                    local_path,
                    'module-parent-master-20170816080815/results'),
            ]
            assert set(result) == set(expected)

    def test_resolve_requires(self):
        build = models.ModuleBuild.query.get(2)
        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='db')
        result = resolver.resolve_requires([":".join([
            build.name, build.stream, build.version, build.context])])

        assert result == {
            'testmodule': {
                'stream': 'master', 'version': '20170109091357', 'context': u'78e4a6fd',
                'ref': 'ff1ea79fc952143efeed1851aa0aa006559239ba',
                'koji_tag': 'module-testmodule-master-20170109091357-78e4a6fd'
            }}

    def test_resolve_profiles(self):
        """
        Tests that the profiles get resolved recursively
        """
        mmd = models.ModuleBuild.query.get(2).mmd()
        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='db')
        result = resolver.resolve_profiles(mmd, ('buildroot', 'srpm-buildroot'))
        expected = {
            'buildroot':
                set(['unzip', 'tar', 'cpio', 'gawk', 'gcc', 'xz', 'sed',
                     'findutils', 'util-linux', 'bash', 'info', 'bzip2',
                     'grep', 'redhat-rpm-config', 'fedora-release',
                     'diffutils', 'make', 'patch', 'shadow-utils', 'coreutils',
                     'which', 'rpm-build', 'gzip', 'gcc-c++']),
            'srpm-buildroot':
                set(['shadow-utils', 'redhat-rpm-config', 'rpm-build',
                     'fedora-release', 'fedpkg-minimal', 'gnupg2',
                     'bash'])
        }
        assert result == expected

    @patch("module_build_service.config.Config.system",
           new_callable=PropertyMock, return_value="test")
    @patch("module_build_service.config.Config.mock_resultsdir",
           new_callable=PropertyMock,
           return_value=os.path.join(base_dir, 'staged_data', "local_builds"))
    def test_resolve_profiles_local_module(self, local_builds, conf_system):
        """
        Test that profiles get resolved recursively on local builds
        """
        with app.app_context():
            utils.load_local_builds(['platform'])
            mmd = models.ModuleBuild.query.get(2).mmd()
            resolver = mbs_resolver.GenericResolver.create(tests.conf, backend='mbs')
            result = resolver.resolve_profiles(mmd, ('buildroot', 'srpm-buildroot'))
            expected = {
                'buildroot':
                    set(['foo']),
                'srpm-buildroot':
                    set(['bar'])
            }
            assert result == expected
